/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.definition;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;

/**
 * @author Tom Baeyens
 */
public abstract class CompositeElement implements Serializable {

  private static final long serialVersionUID = 1L;

  protected long dbid;
  protected int dbversion;
  protected String name;
  protected String description;

  protected List<Node> nodes;

  protected transient Map<String, Node> nodesMap;


  protected BpelProcess processDefinition;
  protected ExceptionHandler exceptionHandler;

  // nested nodes /////////////////////////////////////////////////////////////

  /**
   * creates a nested node with the given name. Also the nested node's parent
   * pointer will be set appropriatly.
   *
   * @param nodeUUID
   *          may be null.
   */
  public Node createNode(final ActivityDefinitionUUID nodeUUID) {
    final Node node = new Node();
    node.setName(nodeUUID.toString());
    this.addNode(node);
    return node;
  }

  public Node addNode(final Node node) {
    node.setProcessDefinition(this.processDefinition);
    if (this.nodes == null) {
      this.nodes = new ArrayList<Node>();
    }
    if (!this.nodes.contains(node)) {
      this.nodes.add(node);
    }
    this.nodesMap = null;
    return node;
  }

  /**
   * removes the given node from the nested activities. Also the node's parent
   * will be nulled. This method will do nothing if the node is null or if the
   * node is not in the list of nested activities. If the node is actually
   * removed from the list of activities, the node's source will be nulled. In
   * case this is the node that was in the activitiesMap and another node exists
   * with the same name, that node (the first) will be put in the activitiesMap
   * as a replacement for the removed node.
   */
  public boolean removeNode(final Node node) {
    if ((node != null) && (this.nodes != null)) {
      final boolean isRemoved = this.nodes.remove(node);
      if (isRemoved) {
        node.setParentNode(null);
        if (this.nodes.isEmpty()) {
          this.nodes = null;
        }
        this.nodesMap = null;
      }
      return isRemoved;
    }
    return false;
  }

  /**
   * the first composite node with the given name or null of no such node
   * exists.
   */
  public Node getNode(final String nodeName) {
    return this.getNodesMap() != null ? this.nodesMap.get(nodeName) : null;
  }

  /** is this node present ? */
  public boolean hasNode(final String nodeName) {
    return (this.getNodesMap() != null) && (this.nodesMap.containsKey(nodeName));
  }

  /**
   * searches for the given node in this node, including this node and all child
   * nodes recursively. The search is done depth-first. A null value for
   * nodeName matches a node without a name.
   */
  public Node findNode(final String nodeName) {
    if (this.nodes != null) {
      for (final Node n : this.nodes) {
        final Node node = n.findNode(nodeName);
        if (node != null) {
          return node;
        }
      }
    }
    return null;
  }

  /**
   * the list of nested activities. Beware: the actual member is returned. No
   * copy is made.
   */
  public List<Node> getNodes() {
    return this.nodes;
  }

  /**
   * the composite nodes, keyed by node name. If a node with the same name
   * occurs multiple times, the first in the list is included in the map. Nodes
   * with a null value for their name are not included in the map. Beware: the
   * actual member is returned. No copy is made. In fact, the returned map is
   * maintained as a cache. So updates to the map will influence subsequent
   * retrievals of nodes by name.
   */
  public Map<String, Node> getNodesMap() {
    if (this.nodesMap == null) {
      this.nodesMap = Node.getNodesMap(this.getNodes());
    }
    return this.nodesMap;
  }

  /** indicates if this processDefinition has nodes. */
  public boolean hasNodes() {
    return (this.nodes != null) && (!this.nodes.isEmpty());
  }

  // getters and setters //////////////////////////////////////////////////////

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public String getDescription() {
    return this.description;
  }

  public void setDescription(final String description) {
    this.description = description;
  }

  /** the database id. */
  public long getDbid() {
    return this.dbid;
  }

  public BpelProcess getProcessDefinition() {
    return this.processDefinition;
  }

  public void setProcessDefinition(final BpelProcess processDefinition) {
    this.processDefinition = processDefinition;
  }

  public ExceptionHandler getExceptionHandler() {
    return this.exceptionHandler;
  }

  public void setExceptionHandler(final ExceptionHandler exceptionHandler) {
    this.exceptionHandler = exceptionHandler;
  }

  public CompositeElement getParent() {
    return null;
  }
}
