/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.definition;

import java.io.Serializable;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.runtime.BpelExecution;
import org.ow2.orchestra.runtime.op.MoveToChildNode;
import org.ow2.orchestra.services.commands.CommandService;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.ReflectUtil;

/**
 * @author Tom Baeyens
 * @author Guillaume Porcher
 */
public class ExceptionHandler implements Serializable {

  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(ExceptionHandler.class.getName());

  protected long dbid;
  protected int dbversion;
  protected String exceptionClassName;
  protected boolean isTransactional = false;
  protected boolean isRethrowMasked = false;
  protected String nodeName;

  // runtime behaviour methods ////////////////////////////////////////////////

  public boolean matches(final Exception exception) {
    return this.matches(exception.getClass());
  }

  public boolean matches(final Class< ? > exceptionClass) {
    if (exceptionClass == null) {
      return false;
    }
    final Class< ? > exceptionClassClass = ReflectUtil.loadClass(Thread.currentThread().getContextClassLoader(),
      this.exceptionClassName);
    return exceptionClassClass.isAssignableFrom(exceptionClass);
  }

  /**
   * Handles the exception. This method will execute compensating code
   * in the same TX or in another TX (depending on the value of {@link #isTransactional} field).
   * The compensating code is not in this method,
   * but in {@link #executeHandler(ExecutionImpl, Exception, ProcessElementImpl)} method.
   * @param execution the execution which produced the exception
   * @param exception the exception to handle
   * @param processElementImpl the process element the exception handler was triggered from.
   */
  public void handle(final BpelExecution execution, final Exception exception, final CompositeElement processElementImpl) {
    if (this.isTransactional) {
      final Environment environment = Environment.getCurrent();
      final Transaction transaction = environment != null ? environment.get(Transaction.class) : null;
      if (transaction != null) {
        Misc.fastDynamicLog(ExceptionHandler.LOG, Level.FINEST, "registering exception handler to %s", transaction);
        final CommandService commandService = environment.get(CommandService.class);
        if (commandService == null) {
          throw new OrchestraRuntimeException(
              "environment doesn't have a command service for registering transactional exception handler",
              exception);
        }
        final ExceptionHandlerSynchronization exceptionHandlerSynchronization = new ExceptionHandlerSynchronization(
            this, execution, exception, processElementImpl, commandService);
        // registration of the synchronization is delegated to the
        // AfterTxCompletionListener
        // to avoid a dependency on class Synchronization
        exceptionHandlerSynchronization.register(transaction);
        Misc.fastDynamicLog(ExceptionHandler.LOG, Level.FINEST, "registering exception handler to %s", transaction);
        throw new OrchestraRuntimeException(
            "transaction exception handler registered handler after transaction completed.  "
            + "make sure this transaction is rolled back",
            exception);
      } else {
        throw new OrchestraRuntimeException(
            "no transaction present in the environment for transactional exception handler",
            exception);
      }
    } else {
      this.executeHandler(execution, exception, processElementImpl);
    }
  }

  /**
   * Compensating code for the exception.
   * This method is called from {@link #handle(ExecutionImpl, Exception, ProcessElementImpl)}
   * or from {@link ExceptionHandlerSynchronization} and should not be called directly.
   *
   * Extensions should overwrite this method.
   *
   * @param execution the execution which produced the exception
   * @param exception the exception to handle
   * @param processElementImpl the process element the exception handler was triggered from.
   */
  protected void executeHandler(final BpelExecution execution, final Exception exception,
      final CompositeElement processElementImpl) {

    if (this.nodeName != null) {
      // perform atomic operation is only offering a new Operation.
      // execution of operation is not done when the method is returning.
      // as a result, we need to explicitly stop execution propagation
      // to be sure exception node has been executed.
      // Normal execution has to be resumed (if wanted) manually.
      execution.waitForSignal();
    }
    // execute child node of process element
    Node childNode = null;
    if (processElementImpl != null) {
      childNode = processElementImpl.getNode(this.nodeName);
    }
    if (childNode != null) {
      Misc.fastDynamicLog(ExceptionHandler.LOG, Level.FINEST, "%s executes node %s", this, this.nodeName);
      execution.performAtomicOperation(new MoveToChildNode(childNode));
    } else {
      Misc.fastDynamicLog(ExceptionHandler.LOG, Level.WARNING,
          "WARNING: %s couldn't find child node %s", this, this.nodeName);
    }
  }

  public static void rethrow(final Exception exception, final String prefixMessage) {
    Misc.fastDynamicLog(ExceptionHandler.LOG, Level.FINEST, "rethrowing %s", exception);
    if (exception instanceof RuntimeException) {
      throw (RuntimeException) exception;
    } else {
      throw new OrchestraRuntimeException(prefixMessage + ": " + exception.getMessage(),
          exception);
    }
  }

  @Override
  public String toString() {
    return this.exceptionClassName != null
      ? "exception-handler(" + this.exceptionClassName + ")" : "exception-handler";
  }

  // getters and setters //////////////////////////////////////////////////////

  public long getDbid() {
    return this.dbid;
  }

  public String getExceptionClassName() {
    return this.exceptionClassName;
  }

  public void setExceptionClassName(final String exceptionClassName) {
    this.exceptionClassName = exceptionClassName;
  }

  public boolean isTransactional() {
    return this.isTransactional;
  }

  public void setTransactional(final boolean isTransactional) {
    this.isTransactional = isTransactional;
  }

  public String getNodeName() {
    return this.nodeName;
  }

  public void setNodeName(final String nodeName) {
    this.nodeName = nodeName;
  }

  public boolean isRethrowMasked() {
    return this.isRethrowMasked;
  }

  public void setRethrowMasked(final boolean isRethrowMasked) {
    this.isRethrowMasked = isRethrowMasked;
  }

}
