/**
 * Bull SAS / OW2 Orchestra
 * Copyright (C) 2007 Bull S.A.S, and individual contributors as indicated
 * by the @authors tag.
 *
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.deployment;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.extension.ExtensionProvider;
import org.ow2.orchestra.facade.Deployment;
import org.ow2.orchestra.facade.def.ProcessDefinition;
import org.ow2.orchestra.facade.def.full.ProcessFullDefinition;
import org.ow2.orchestra.facade.exception.DeploymentException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.facade.wsdl.WsdlDefinitions;
import org.ow2.orchestra.parsing.BpelParser;
import org.ow2.orchestra.parsing.binding.DefinitionKeeper;
import org.ow2.orchestra.parsing.binding.extension.ExtensionsInfo;
import org.ow2.orchestra.runtime.BpelInstance;
import org.ow2.orchestra.services.OperationKey;
import org.ow2.orchestra.services.PendingMessage;
import org.ow2.orchestra.services.Receiver;
import org.ow2.orchestra.services.ResultIterator;
import org.ow2.orchestra.services.handlers.UndeployedProcessHandler;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.Namespaces;
import org.ow2.orchestra.util.ProcessResourcesRepository;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.ow2.orchestra.xml.Parse;
import org.xml.sax.InputSource;

import javax.xml.namespace.QName;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * takes {@link Deployment}s as input, parses the contents into a
 * {@link BpelProcess}, stores the {@link BpelProcess} into the
 * {@link Repository} and publishes the services with a {@link org.ow2.orchestra.services.itf.Publisher}.
 * <p/>
 * <p>
 * processRepository and servicePublisher have to be injected by the user (or an
 * IoC framework) before the deploy method is invoked.
 * </p>
 * <p/>
 * <p>
 * This class is thread safe and can be shared between multiple threads.
 * </p>
 *
 * @author Tom Baeyens, Goulven Le Jeune
 */
public class Deployer {

  private static final Logger LOG = Logger.getLogger(Deployer.class.getName());

  /**
   * parser to use to parse bpel files.
   */
  private BpelParser bpelParser = null;

  /**
   * Default constructor.
   */
  public Deployer() {
    this.bpelParser = new BpelParser();
  }

  /**
   * Thanks to the deployment object given in parameter, parse all wsdl files,
   * bpel file. Then store the process in processRepository and publish bpel
   * process services.
   *
   * @param bpelDeployment populated bpelDeployment
   * @param environment    environment
   * @return created BpelProcess
   */
  public ProcessFullDefinition deploy(final Deployment bpelDeployment, final Environment environment) {
    Misc.checkArgsNotNull(bpelDeployment, environment);

    Misc.fastDynamicLog(Deployer.LOG, Level.INFO, "Deploying BPEL deployment %s", bpelDeployment);

    // now the bpel process will be parsed.
    final InputSource bpel = bpelDeployment.getBpel();
    if (bpel == null) {
      throw new OrchestraRuntimeException("Unable to open BPEL Stream");
    }

    final Parse parse = this.bpelParser.createParse();

    final WsdlDefinitions wsdlDefinitions = bpelDeployment.getWsdlDefinitions();
    parse.pushObject(wsdlDefinitions);

    final Namespaces namespaces = new Namespaces(parse);
    parse.pushObject(namespaces);

    ProcessFullDefinition processDefinition = null;
    final ProcessResourcesRepository resourcesRepository = bpelDeployment.getResourcesRepository();
    parse.pushObject(resourcesRepository);

    final WsdlsInfos wsdlInfos = new WsdlsInfos();
    parse.pushObject(wsdlInfos);

    final ExtensionsInfo extensionsInfos = new ExtensionsInfo();
    parse.pushObject(extensionsInfos);

    BpelProcess bpelProcess = null;

    final Repository repository = environment.get(Repository.class);

    // make the bpel process available to the bpel parser by pushing it
    // on the object stack in the parse.
    final DefinitionKeeper definitionKeeper = new DefinitionKeeper(repository.getNextProcessCount());
    parse.pushObject(definitionKeeper);

    try {
      bpelProcess = (BpelProcess) parse.setInputSource(bpel).execute().getDocumentObject();
    } finally {
      // remove defKeeper
      parse.popObject();
      // remove wsdlInfos
      parse.popObject();
      // remove extensionsInfos
      parse.popObject();
      // remove styleSheetRepository
      parse.popObject();
      // remove namespaces
      parse.popObject();
      // remove wsdlDefinitions
      parse.popObject();
    }
    if (!parse.hasProblems()) {
      if (bpelProcess == null) {
        throw new OrchestraRuntimeException("Bpel Process null");
      }
      bpelProcess.staticAnalysisCheck(parse);
    }
    parse.checkProblems("bpel file");

    final QName processQName = bpelProcess.getQName();
    // Get other versions of the process
    final List<BpelProcess> oldBpelProcesses = repository.getProcesses(processQName);
    int maxVersion = 0;
    // For each old version, set the old version retired
    // Find the max version deployed
    for (final BpelProcess oldBpelProcess : oldBpelProcesses) {
      repository.retireProcess(oldBpelProcess);
      if (oldBpelProcess.getVersion() > maxVersion) {
        maxVersion = oldBpelProcess.getVersion();
      }
    }
    // set the new process version
    bpelProcess.setVersion(maxVersion + 1);

    // store new process in repository
    repository.storeProcess(bpelProcess);

    // deploy extensions
    final List<ExtensionProvider> extensionProviders = EnvTool.get(List.class, "extensionProviders");
    if (extensionProviders != null) {
      for (ExtensionProvider extensionProvider : extensionProviders) {
        extensionProvider.deployExternalModule(bpelProcess);
      }
    }

    // Record process deployed
    processDefinition = definitionKeeper.getProcessFullDefinition();
    processDefinition.setVersion(bpelProcess.getVersion());

    EnvTool.getRecorder().recordProcessDeployed(processDefinition);

    Misc.fastDynamicLog(Deployer.LOG, Level.INFO, "BPEL process %s deployed.", processQName);
    return processDefinition;
  }

  public void undeploy(final QName processQName, final Environment environment) {
    final Repository repository = environment.get(Repository.class);
    final BpelProcess bpelProcess = repository.getActiveProcess(processQName);

    this.undeploy(bpelProcess.getUUID(), environment);

    Misc.fastDynamicLog(Deployer.LOG, Level.INFO, "BPEL process %s undeployed.", processQName);
  }

  public void undeploy(final ProcessDefinitionUUID pduuid, final Environment environment) {
    final Repository repository = environment.get(Repository.class);
    final BpelProcess bpelProcess = repository.getProcess(pduuid);
    if (bpelProcess == null) {
      throw new DeploymentException("Problem during undeployment : process " + pduuid + " not found");
    }
    // check no more instances are running
    final List<BpelInstance> runningInstances = repository.getInstances(pduuid);
    if (runningInstances != null && !runningInstances.isEmpty()) {
      throw new DeploymentException(
              "Problem during undeployment: some instances  of process " + pduuid + " are still running");
    }
    final boolean isActive = bpelProcess.equals(EnvTool.getRepository().getActiveProcess(bpelProcess.getQName()));
    // check other versions of the process
    final List<BpelProcess> processVersions = repository.getProcesses(bpelProcess.getQName());
    processVersions.remove(bpelProcess);

    if (!processVersions.isEmpty() && isActive) {
      // Undeployed process was active.
      // Get the highest version of the process.
      BpelProcess highestProcessVersion = null;
      for (final BpelProcess otherProcessVersion : processVersions) {
        if (highestProcessVersion == null || highestProcessVersion.getVersion() < otherProcessVersion.getVersion()) {
          highestProcessVersion = otherProcessVersion;
        }
      }
      // Set highest version as active.
      if (highestProcessVersion != null) {
        repository.activateProcess(highestProcessVersion);
      }
    }


    // Remove old pending messages
    this.removeUndeployedProcessPendingMessages(environment, bpelProcess, processVersions);

    // remove process from repository
    repository.removeProcess(pduuid);

    // Record process undeployed
    EnvTool.getRecorder().recordProcessUndeployed(bpelProcess.getUUID());

    // call undeployed process handler.
    final ProcessDefinition processDefinition = EnvTool.getQuerier().getProcessDefinition(bpelProcess.getUUID());
    final UndeployedProcessHandler handler = EnvTool.get(UndeployedProcessHandler.class, UndeployedProcessHandler.DEFAULT_KEY);
    handler.handleUndeployedProcess(processDefinition);
    Misc.fastDynamicLog(Deployer.LOG, Level.INFO, "BPEL process %s undeployed.", pduuid);

  }

  /**
   * Removes pending messages associated with this process
   * if no other versions of the process share the operationKey of the message.
   *
   * @param environment
   * @param bpelProcess     process undeployed
   * @param processVersions other deployed versions of the undeployed process
   */
  private void removeUndeployedProcessPendingMessages(
          final Environment environment,
          final BpelProcess bpelProcess,
          final List<BpelProcess> processVersions) {
    // Get publisher
    final Publisher publisher = environment.get(Publisher.class);
    // Get operation keys of the process
    final Set<OperationKey> processOperationKeys = publisher.getServiceOperationKeys(bpelProcess, environment);
    // For all other versions
    for (final BpelProcess otherVersion : processVersions) {
      // Get operation keys of the other version
      final Set<OperationKey> otherVersionOpKeys = publisher.getServiceOperationKeys(otherVersion, environment);
      // Remove operation keys used by the other version
      processOperationKeys.removeAll(otherVersionOpKeys);
    }
    // processOperationKeys contains only operation keys that are not used by any other version of the process.
    for (final OperationKey opKey : processOperationKeys) {
      // Remove all pending message associated with the operation key to remove.
      ResultIterator<PendingMessage> pendingMessages = EnvTool.getRepository().getAllPendingMessages(opKey);
      PendingMessage pendingMessage = pendingMessages.next(false); // do not evict from cache because object is modified (deleted)
      while (pendingMessage != null) {
        Receiver.deletePendingMessage(pendingMessage);
        pendingMessage = pendingMessages.next(false); // do not evict from cache because object is modified (deleted)
      }
    }
  }
}
