/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.env;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.Misc;

/**
 * @author Tom Baeyens
 */
public class BasicEnvironment extends Environment {

  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(BasicEnvironment.class.getName());

  protected String userId;
  protected Map<String, Context> contexts;
  protected ArrayList<String> defaultSearchOrderList;
  protected String[] defaultSearchOrder;
  protected Throwable exception;

  protected transient ClassLoader classLoader;

  public BasicEnvironment() {
    this.contexts = new HashMap<String, Context>();
    this.defaultSearchOrderList = new ArrayList<String>();
    this.defaultSearchOrder = null;
    Environment.pushEnvironment(this);
  }

  // context methods
  // ////////////////////////////////////////////////////////////

  @Override
  public Context getContext(final String contextName) {
    return this.contexts.get(contextName);
  }

  @Override
  public void addContext(final Context context) {
    final String key = context.getName();
    this.contexts.put(key, context);
    this.defaultSearchOrderList.add(key);
    this.defaultSearchOrder = null;
  }

  @Override
  public void removeContext(final Context context) {
    final String contextName = context.getName();
    final Context removedContext = this.contexts.remove(contextName);
    if (removedContext != null) {
      this.defaultSearchOrderList.remove(contextName);
      this.defaultSearchOrder = null;
    }
  }

  public Context getEnvironmentFactoryContext() {
    return this.getContext(Context.CONTEXTNAME_ENVIRONMENT_FACTORY);
  }

  public Context getEnvironmentContext() {
    return this.getContext(Context.CONTEXTNAME_ENVIRONMENT);
  }

  // userId methods ///////////////////////////////////////////////////////////

  @Override
  public String getUserId() {
    // if the authenticated user was explicitely set
    if (this.userId != null) {
      // return that one
      return this.userId;
    }

    // if an Authentication was specified
    final Authentication authentication = this.get(Authentication.class);
    if (authentication != null) {
      // let the authentication do the work
      return authentication.getUserId();
    }

    return null;
  }

  @Override
  public void setUserId(final String userId) {
    this.userId = userId;
  }

  // classloader methods //////////////////////////////////////////////////////

  @Override
  public ClassLoader getClassLoader() {
    return this.classLoader;
  }

  @Override
  public void setClassLoader(final ClassLoader classLoader) {
    this.classLoader = classLoader;
  }

  // search methods ///////////////////////////////////////////////////////////

  @Override
  public Object get(final String name) {
    return this.get(name, null);
  }

  @Override
  public Object get(final String name, String[] searchOrder) {
    if (searchOrder == null) {
      searchOrder = this.getDefaultSearchOrder();
    }
    for (final String element : searchOrder) {
      final Context context = this.contexts.get(element);
      if (context.has(name)) {
        return context.get(name);
      }
    }
    return null;
  }

  @Override
  public <T> T get(final Class<T> type) {
    return this.get(type, null);
  }

  @Override
  public <T> T get(final Class<T> type, String[] searchOrder) {
    if (searchOrder == null) {
      searchOrder = this.getDefaultSearchOrder();
    }
    for (final String element : searchOrder) {
      final Context context = this.contexts.get(element);
      final T o = context.get(type);
      if (o != null) {
        return o;
      }
    }
    return null;
  }

  // close ////////////////////////////////////////////////////////////////////

  @Override
  public void close() {
    Misc.fastDynamicLog(BasicEnvironment.LOG, Level.FINEST, "closing %s", this);

    final Environment popped = Environment.popEnvironment();
    if (this != popped) {
      throw new OrchestraRuntimeException("environment nesting problem");
    }
  }

  // private methods //////////////////////////////////////////////////////////

  protected String[] getDefaultSearchOrder() {
    if (this.defaultSearchOrder == null) {
      final int size = this.defaultSearchOrderList.size();
      this.defaultSearchOrder = new String[size];
      for (int i = 0; i < size; i++) {
        this.defaultSearchOrder[i] = this.defaultSearchOrderList.get(size - 1 - i);
      }
    }
    return this.defaultSearchOrder;
  }
}
