package org.ow2.orchestra.env;

import org.ow2.orchestra.env.xml.EnvironmentFactoryParser;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.db.DbHistory;
import org.ow2.orchestra.persistence.db.DbJournal;
import org.ow2.orchestra.persistence.log.LoggerArchiver;
import org.ow2.orchestra.persistence.log.LoggerRecorder;
import org.ow2.orchestra.services.Archiver;
import org.ow2.orchestra.services.Querier;
import org.ow2.orchestra.services.Recorder;
import org.ow2.orchestra.services.handlers.FinishedInstanceHandler;
import org.ow2.orchestra.services.handlers.UndeployedProcessHandler;
import org.ow2.orchestra.services.handlers.impl.NoOpFinishedInstanceHandler;
import org.ow2.orchestra.services.handlers.impl.NoOpUndeployedProcessHandler;
import org.ow2.orchestra.services.impl.DbRepository;
import org.ow2.orchestra.services.impl.PublisherImpl;
import org.ow2.orchestra.services.impl.UnsupportedInvoker;
import org.ow2.orchestra.services.itf.Invoker;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.services.jobexecutor.DeadJobHandler;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.ReflectUtil;
import org.ow2.orchestra.util.XmlUtil;
import org.w3c.dom.Document;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Generates a default environment configuration for running Orchestra.
 *
 * @author Guillaume Porcher, Charles Souillard
 */
public class DefaultEnvXMLGenerator {

  public static final String CORE_HB_CONF_NAME = "hibernate-configuration:core";
  public static final String CORE_HB_SESSION_FACTORY_NAME = "hibernate-session-factory:core";
  public static final String CORE_HB_SESSION_NAME = "hibernate-session:core";
  public static final String CORE_RUNTIME_SESSION_NAME = "runtime-session:core";
  public static final String CORE_QUERIER_SESSION_NAME = "querier-session:core";

  public static final String JOURNAL_HB_CONF_NAME = "hibernate-configuration:journal";
  public static final String JOURNAL_HB_SESSION_FACTORY_NAME = "hibernate-session-factory:journal";
  public static final String JOURNAL_HB_SESSION_NAME = "hibernate-session:journal";
  public static final String JOURNAL_QUERIER_SESSION_NAME = "querier-session:journal";

  public static final String HISTORY_HB_CONF_NAME = "hibernate-configuration:history";
  public static final String HISTORY_HB_SESSION_FACTORY_NAME = "hibernate-session-factory:history";
  public static final String HISTORY_HB_SESSION_NAME = "hibernate-session:history";
  public static final String HISTORY_QUERIER_SESSION_NAME = "querier-session:history";

  public static final String ORCHESTRA_PROPERTIES = "orchestra-properties";

  private final Map<String, String> applicationVariables;

  private final Map<String, String> blockVariables;

  private Class<? extends Repository> repositoryType = null;

  private Class<? extends Publisher> publisherType = null;

  private Class<? extends Invoker> invokerType = null;

  private List<String> extensionProviders = new ArrayList<String>();

  private String recorderXml = "";
  private String archiverXml = "";
  private String querierXml = "";

  private String fihXml = "";
  private String uphXml = "";
  private String deadJobHandlerXml = "";


  public static enum InstallationType {
    NO_JOURNAL(false, false, false),
    INTERNAL_JOURNAL_WITH_HISTORY(true, true, true),
    EXTERNAL_JOURNAL_WITH_HISTORY(true, true, false),
    INTERNAL_JOURNAL_NO_HISTORY(true, false, true),
    EXTERNAL_JOURNAL_NO_HISTORY(true, false, false);

    private final boolean hasJournal;
    private final boolean hasHistory;
    private final boolean journalUsesCoreSession;

    private InstallationType(final boolean hasJournal, final boolean hasHistory, final boolean journalUsesCoreSession) {
      this.hasJournal = hasJournal;
      this.hasHistory = hasHistory;
      this.journalUsesCoreSession = journalUsesCoreSession;
    }

    /**
     * @return the hasJournal
     */
    public boolean hasJournal() {
      return this.hasJournal;
    }

    /**
     * @return the hasHistory
     */
    public boolean hasHistory() {
      return this.hasHistory;
    }
  }

  /**
   * Creates an environment generator
   *
   * @param hibernateRelativePath if <code>null</code>, hibernate.properties will be referenced as a resource.
   *                              if not null, it will be referenced as a file, with path ${hibernateRelativePath}/hibernate.properties
   */
  public DefaultEnvXMLGenerator(final String hibernateRelativePath, final InstallationType installationType) {
    this.applicationVariables = new LinkedHashMap<String, String>();
    this.blockVariables = new LinkedHashMap<String, String>();
    this.setPublisherType(PublisherImpl.class);
    this.setInvokerType(UnsupportedInvoker.class);

    /**
     * Orchestra.properties
     */
    this.setApplicationEntry(DefaultEnvXMLGenerator.ORCHESTRA_PROPERTIES,
            "<properties name='" + DefaultEnvXMLGenerator.ORCHESTRA_PROPERTIES + "' resource='orchestra.properties' />");
    /*
     * Runtime (Engine + journal) Hibernate configuration
     */
    String hibernateCoreConfigXml = "";
    hibernateCoreConfigXml += "<hibernate-configuration name='" + DefaultEnvXMLGenerator.CORE_HB_CONF_NAME + "'>";
    if (hibernateRelativePath == null || hibernateRelativePath.length() == 0) {
      hibernateCoreConfigXml += "<properties resource='hibernate.properties' />";
    } else {
      hibernateCoreConfigXml += "<properties file='" + hibernateRelativePath + "hibernate.properties' />";
    }
    hibernateCoreConfigXml += "<mappings resource='hibernate/bpel.core.mappings.xml' />";
    if (installationType.journalUsesCoreSession) {
      hibernateCoreConfigXml += "<mappings resource='hibernate/bpel.monitoring.mappings.xml' />";
    }
    hibernateCoreConfigXml +=
            "<cache-configuration resource='hibernate/bpel.cache.definition.xml' usage='nonstrict-read-write' />";
    hibernateCoreConfigXml +=
            "<!-- cache-configuration for runtime data. Must be used with a replicated cache when used in cluster -->";
    hibernateCoreConfigXml +=
            "<!-- <cache-configuration resource='hibernate/bpel.cache.runtime.xml' usage='read-write' /> -->";
    hibernateCoreConfigXml += "</hibernate-configuration>";

    this.setApplicationEntry(DefaultEnvXMLGenerator.CORE_HB_CONF_NAME, hibernateCoreConfigXml);
    this.setApplicationEntry(DefaultEnvXMLGenerator.CORE_HB_SESSION_FACTORY_NAME,
            "<hibernate-session-factory name='" + DefaultEnvXMLGenerator.CORE_HB_SESSION_FACTORY_NAME
                    + "' configuration='" + DefaultEnvXMLGenerator.CORE_HB_CONF_NAME + "' init='eager' />");

    this.setBlockEntry(DefaultEnvXMLGenerator.CORE_HB_SESSION_NAME,
            "<hibernate-session name='" + DefaultEnvXMLGenerator.CORE_HB_SESSION_NAME
                    + "' factory='" + DefaultEnvXMLGenerator.CORE_HB_SESSION_FACTORY_NAME + "' />");

    this.setBlockEntry(DefaultEnvXMLGenerator.CORE_RUNTIME_SESSION_NAME,
            "<runtime-db-session name='" + DefaultEnvXMLGenerator.CORE_RUNTIME_SESSION_NAME
                    + "' session='" + DefaultEnvXMLGenerator.CORE_HB_SESSION_NAME + "'/>");

    /*
     * Other services
     */
    String commandServiceConfigXml = "<command-service>";
    commandServiceConfigXml += "<retry-interceptor retries='10' delay-factor='4' delay='100' max-delay='10000' />";
    commandServiceConfigXml += "<environment-interceptor />";
    commandServiceConfigXml += "<standard-transaction-interceptor />";
    commandServiceConfigXml += "</command-service>";
    this.setApplicationEntry("command-service", commandServiceConfigXml);
    this.setApplicationEntry("invoke-executor", "<invoke-executor threads='10' />");
    this.setApplicationEntry("job-executor",
            "<job-executor threads='10' auto-start='false' lock='180000' limit-job-per-instance='false' acquireSize='1' addQueueSize='20' queueSize='20' />");

    // replace previous line by this one to use jdk job executor
    // this.setApplicationEntry("job-executor", "    <job-executor auto-start='false' type='jdk' />");

    this.setRepositoryType(DbRepository.class);

    this.setBlockEntry("transaction", "<transaction />");

    this.setBlockEntry("timer_session", "<timer-session retries='50' />");
    this.setBlockEntry("message_session", "<message-session retries='50' use-fair-scheduling='false' />");
    this.setBlockEntry("job-session", "<job-db-session session='" + DefaultEnvXMLGenerator.CORE_HB_SESSION_NAME + "' />");


    this.setRecorderType(LoggerRecorder.class);

    this.setFinishedInstanceHandlerType(NoOpFinishedInstanceHandler.class);
    this.setUndeployedProcessHandlerType(NoOpUndeployedProcessHandler.class);

    this.setArchiverType(LoggerArchiver.class);

    if (installationType.hasJournal) {
      this.addJournal(hibernateRelativePath, installationType.journalUsesCoreSession);
    }

    if (installationType.hasHistory) {
      this.addHistory(hibernateRelativePath);
    }
  }


  private void addJournal(final String hibernateRelativePath, final boolean journalUsesCoreSession) {
    /*
     * Journal creation (CORE_QUERIER_SESSION is given to constructor)
     */
    if (journalUsesCoreSession) {
      this.setBlockEntry("journal", "<journal name='journal' class='" + DbJournal.class.getName() + "'>"
              + "<arg><ref object='" + DefaultEnvXMLGenerator.CORE_QUERIER_SESSION_NAME + "' /></arg></journal>");

      this.setBlockEntry(DefaultEnvXMLGenerator.CORE_QUERIER_SESSION_NAME,
              "<querier-db-session name='" + DefaultEnvXMLGenerator.CORE_QUERIER_SESSION_NAME
                      + "' session='" + DefaultEnvXMLGenerator.CORE_HB_SESSION_NAME + "'/>");

    } else {
      /*
       * Journal Hibernate configuration
       */
      String hibernateJournalConfigXml = "";
      hibernateJournalConfigXml += "<hibernate-configuration name='" + DefaultEnvXMLGenerator.JOURNAL_HB_CONF_NAME + "'>";
      if (hibernateRelativePath == null || hibernateRelativePath.length() == 0) {
        hibernateJournalConfigXml += "<properties resource='hibernate-journal.properties' />";
      } else {
        hibernateJournalConfigXml += "<properties file='" + hibernateRelativePath + "hibernate-journal.properties' />";
      }

      hibernateJournalConfigXml += "<mappings resource='hibernate/bpel.monitoring.mappings.xml' />";
      hibernateJournalConfigXml += "</hibernate-configuration>";

      this.setApplicationEntry(DefaultEnvXMLGenerator.JOURNAL_HB_CONF_NAME, hibernateJournalConfigXml);
      this.setApplicationEntry(DefaultEnvXMLGenerator.JOURNAL_HB_SESSION_FACTORY_NAME,
              "<hibernate-session-factory name='" + DefaultEnvXMLGenerator.JOURNAL_HB_SESSION_FACTORY_NAME
                      + "' configuration='" + DefaultEnvXMLGenerator.JOURNAL_HB_CONF_NAME + "' init='eager' />");

      this.setBlockEntry(DefaultEnvXMLGenerator.JOURNAL_HB_SESSION_NAME,
              "<hibernate-session name='" + DefaultEnvXMLGenerator.JOURNAL_HB_SESSION_NAME
                      + "' factory='" + DefaultEnvXMLGenerator.JOURNAL_HB_SESSION_FACTORY_NAME + "' />");

      this.setBlockEntry(DefaultEnvXMLGenerator.JOURNAL_QUERIER_SESSION_NAME,
              "<querier-db-session name='" + DefaultEnvXMLGenerator.JOURNAL_QUERIER_SESSION_NAME
                      + "' session='" + DefaultEnvXMLGenerator.JOURNAL_HB_SESSION_NAME + "'/>");

      this.setBlockEntry("journal", "<journal name='journal' class='" + DbJournal.class.getName() + "'>"
              + "<arg><ref object='" + DefaultEnvXMLGenerator.JOURNAL_QUERIER_SESSION_NAME + "' /></arg></journal>");
    }

    this.addRecorderType("<ref object='journal' />");
    this.addQuerierType("<ref object='journal' />");

  }

  public void addExtension(String extensionPath) {

    String hibernateExtensionConfigXml = this.applicationVariables.get(DefaultEnvXMLGenerator.CORE_HB_CONF_NAME).replaceAll("<mappings resource='hibernate/bpel.core.mappings.xml' />",
            "<mappings resource='hibernate/bpel.core.mappings.xml' />" + "<mappings resource='" + extensionPath + "' />");
    this.setApplicationEntry(DefaultEnvXMLGenerator.CORE_HB_CONF_NAME, hibernateExtensionConfigXml);
  }


  private void addHistory(final String hibernateRelativePath) {
    /*
     * History creation (HISTORY_QUERIER_SESSION is given to constructor)
     */
    this.setBlockEntry("history", "<history name='history' class='" + DbHistory.class.getName() + "'>"
            + "<arg><ref object='" + DefaultEnvXMLGenerator.HISTORY_QUERIER_SESSION_NAME + "' /></arg></history>");

    this.addArchiverType("<ref object='history' />");
    this.addQuerierType("<ref object='history' />");

    /*
     * History Hibernate configuration
     */
    String hibernateHistoryConfigXml = "";
    hibernateHistoryConfigXml += "<hibernate-configuration name='" + DefaultEnvXMLGenerator.HISTORY_HB_CONF_NAME + "'>";
    if (hibernateRelativePath == null || hibernateRelativePath.length() == 0) {
      hibernateHistoryConfigXml += "<properties resource='hibernate-history.properties' />";
    } else {
      hibernateHistoryConfigXml += "<properties file='" + hibernateRelativePath + "hibernate-history.properties' />";
    }

    hibernateHistoryConfigXml += "<mappings resource='hibernate/bpel.monitoring.mappings.xml' />";
    hibernateHistoryConfigXml += "</hibernate-configuration>";

    this.setApplicationEntry(DefaultEnvXMLGenerator.HISTORY_HB_CONF_NAME, hibernateHistoryConfigXml);
    this.setApplicationEntry(DefaultEnvXMLGenerator.HISTORY_HB_SESSION_FACTORY_NAME,
            "<hibernate-session-factory name='" + DefaultEnvXMLGenerator.HISTORY_HB_SESSION_FACTORY_NAME
                    + "' configuration='" + DefaultEnvXMLGenerator.HISTORY_HB_CONF_NAME + "' init='eager' />");

    this.setBlockEntry(DefaultEnvXMLGenerator.HISTORY_HB_SESSION_NAME,
            "<hibernate-session name='" + DefaultEnvXMLGenerator.HISTORY_HB_SESSION_NAME
                    + "' factory='" + DefaultEnvXMLGenerator.HISTORY_HB_SESSION_FACTORY_NAME + "' />");

    this.setBlockEntry(DefaultEnvXMLGenerator.HISTORY_QUERIER_SESSION_NAME,
            "<querier-db-session name='" + DefaultEnvXMLGenerator.HISTORY_QUERIER_SESSION_NAME
                    + "' session='" + DefaultEnvXMLGenerator.HISTORY_HB_SESSION_NAME + "'/>");
  }

  /**
   *
   */
  public DefaultEnvXMLGenerator() {
    this(null, InstallationType.INTERNAL_JOURNAL_NO_HISTORY);
  }

  /**
   * Generates a xml for the given environment configuration.
   *
   * @return
   */
  public String createEnvironmentXml() {
    if (this.invokerType == null) {
      throw new OrchestraRuntimeException("Error while creating environment xml : invoker is null");
    }
    if (this.publisherType == null) {
      throw new OrchestraRuntimeException("Error while creating environment xml : publisher is null");
    }
    if (this.repositoryType == null) {
      throw new OrchestraRuntimeException("Error while creating environment xml : repository is null");
    }
    if (this.recorderXml.length() > 0) {
      this.setBlockEntry(Recorder.DEFAULT_KEY,
              "<chainer name='" + Recorder.DEFAULT_KEY + "'>"
                      + this.recorderXml
                      + "</chainer>");
    }

    if (this.archiverXml.length() > 0) {
      this.setBlockEntry(Archiver.DEFAULT_KEY,
              "<chainer name='" + Archiver.DEFAULT_KEY + "'>"
                      + this.archiverXml
                      + "</chainer>");
    }

    this.setBlockEntry(Querier.DEFAULT_KEY,
            "<queryApi name='" + Querier.DEFAULT_KEY + "'>"
                    + this.querierXml
                    + "</queryApi>");

    this.setBlockEntry(FinishedInstanceHandler.DEFAULT_KEY, this.fihXml);
    this.setBlockEntry(UndeployedProcessHandler.DEFAULT_KEY, this.uphXml);

    this.setApplicationEntry("dead-job-handler", this.deadJobHandlerXml);

    final StringBuffer result = new StringBuffer();
    result.append("<environment-definition>");
    result.append("<environment-factory>");
    for (final String entry : this.applicationVariables.values()) {
      result.append(entry);
    }
    result.append("<repository class='" + this.repositoryType.getName() + "' />"
            + "<publisher class='" + this.publisherType.getName() + "' />"
            + "<invoker class='" + this.invokerType.getName() + "' name='" + Invoker.DEFAULT_KEY + "'/>");
    if (this.extensionProviders != null) {
      result.append("<list name='extensionProviders'>");
      for (String extensionProvider: this.extensionProviders) {
        result.append(extensionProvider);
      }
      result.append("</list>");
    }
    // End of environment factory scoped objects

    result.append("</environment-factory><environment>");
    // Start of environment scoped objects
    for (final String entry : this.blockVariables.values()) {
      result.append(entry);
    }
    result.append("</environment></environment-definition>");


    // format XML
    final Document doc = XmlUtil.getDocumentFromString(result.toString(), false);
    return XmlUtil.toString(doc);
  }

  /**
   * Creates a new EnvironmentFactory from the list of entries.
   */
  public EnvironmentFactory createEnvironmentFactory() {
    final EnvironmentFactory environmentFactory = EnvironmentFactoryParser
            .parseEnvironmentFactoryFromXmlString(this.createEnvironmentXml());
    return environmentFactory;
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key        objectName
   * @param classToUse classToUse at runtime
   * @return a string that declares an object to be added in the context.
   */
  private static String getObjectDecl(final String key, final Class<?> classToUse) {
    return "    <object name='" + key + "' class='" + classToUse.getName()
            + "' init='eager'/>";
  }

  /**
   * Return a string that declares an object to be added in the context.
   *
   * @param key        objectName
   * @param classToUse classToUse at runtime
   * @param fieldKeys  fields name
   * @param refKeys    the keys of objects to reference in fields declaration
   * @return a string that declares an object to be added in the context.
   */
  private static String getObjectDeclWithFields(final String key, final Class<?> classToUse,
                                                final String[] fieldKeys, final String[] refKeys) {
    if (fieldKeys.length != refKeys.length) {
      throw new RuntimeException(
              "Error : fieldKeys and refKeys don't have the same number of elements");
    }
    final StringBuffer result = new StringBuffer("<object name='" + key + "' class='"
            + classToUse.getName() + "' init='eager'>");
    for (int i = 0; i < fieldKeys.length; i++) {
      result.append("<field name='" + fieldKeys[i] + "'><ref object='" + refKeys[i] + "' /></field>");
    }
    result.append("</object>");
    return result.toString();
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name name of the entry
   * @param xml  declaration of this entry
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final String xml) {
    return this.applicationVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name       name of the entry
   * @param classToUse implementation
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final Class<?> classToUse) {
    final String xml = DefaultEnvXMLGenerator.getObjectDecl(name, classToUse);
    return this.setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name       name of the entry
   * @param classToUse implementation
   * @param fieldKeys  list of fields to set
   * @param refKeys    value to set to the fields
   * @return previous declaration for this name
   */
  public String setApplicationEntry(final String name, final Class<?> classToUse,
                                    final String[] fieldKeys, final String[] refKeys) {
    final String xml = DefaultEnvXMLGenerator.getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return this.setApplicationEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name name of the entry
   * @param xml  declaration of this entry
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final String xml) {
    return this.blockVariables.put(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name       name of the entry
   * @param classToUse implementation
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final Class<?> classToUse) {
    final String xml = DefaultEnvXMLGenerator.getObjectDecl(name, classToUse);
    return this.setBlockEntry(name, xml);
  }

  /**
   * Add or replace an entry in the environment to generate.
   *
   * @param name       name of the entry
   * @param classToUse implementation
   * @param fieldKeys  list of fields to set
   * @param refKeys    value to set to the fields
   * @return previous declaration for this name
   */
  public String setBlockEntry(final String name, final Class<?> classToUse,
                              final String[] fieldKeys, final String[] refKeys) {
    final String xml = DefaultEnvXMLGenerator.getObjectDeclWithFields(name, classToUse, fieldKeys, refKeys);
    return this.setBlockEntry(name, xml);
  }

  public void setRepositoryType(final Class<? extends Repository> repositoryType) {
    this.repositoryType = repositoryType;
  }

  public void setPublisherType(final Class<? extends Publisher> publisherType) {
    this.publisherType = publisherType;
  }

  public void setInvokerType(final Class<? extends Invoker> invokerType) {
    this.invokerType = invokerType;
  }

  public void addExtensionProviderType(String extensionProviderDeclaration) {
    this.extensionProviders.add(extensionProviderDeclaration);
  }

  public void setRecorderType(final Class<? extends Recorder>... recorderClasses) {
    this.recorderXml = "";
    for (final Class<? extends Recorder> recorderClass : recorderClasses) {
      this.addRecorderType("<recorder class='" + recorderClass.getName() + "' />");
    }
  }

  public void addRecorderType(final String recorderDeclaration) {
    this.recorderXml += recorderDeclaration;
  }

  public void setArchiverType(final Class<? extends Archiver>... archiverClasses) {
    this.archiverXml = "";
    for (final Class<? extends Archiver> archiverClass : archiverClasses) {
      this.addArchiverType("<archiver class='" + archiverClass.getName() + "' />");
    }
  }

  public void addArchiverType(final String archiverDeclaration) {
    this.archiverXml += archiverDeclaration;
  }

  public void addQuerierType(final String querierDeclaration) {
    this.querierXml += querierDeclaration;
  }

  public void setFinishedInstanceHandlerType(final Class<? extends FinishedInstanceHandler>... fihClasses) {
    this.fihXml = "<chainer name='" + FinishedInstanceHandler.DEFAULT_KEY + "'>";
    for (final Class<? extends FinishedInstanceHandler> fihClass : fihClasses) {
      this.fihXml += "<finished-instance-handler class='" + fihClass.getName() + "' />";
    }
    this.fihXml += "</chainer>";
  }

  public void setUndeployedProcessHandlerType(final Class<? extends UndeployedProcessHandler>... uphClasses) {
    this.uphXml = "<chainer name='" + UndeployedProcessHandler.DEFAULT_KEY + "'>";
    for (final Class<? extends UndeployedProcessHandler> uphClass : uphClasses) {
      this.uphXml += "<undeployed-process-handler class='" + uphClass.getName() + "' />";
    }
    this.uphXml += "</chainer>";
  }

  public void setDeadJobHandlerType(final Class<? extends DeadJobHandler>... djhClasses) {
    this.deadJobHandlerXml = "";
    if (djhClasses.length > 1) {
      this.deadJobHandlerXml = "<chainer>";
    }
    for (final Class<? extends DeadJobHandler> djhClass : djhClasses) {
      this.deadJobHandlerXml += "<dead-job-handler class='" + djhClass.getName() + "' />";
    }
    if (djhClasses.length > 1) {
      this.deadJobHandlerXml = "</chainer>";
    }
  }

  /**
   * Prints the default environment on standard output.
   *
   * @param args
   * @throws java.io.IOException
   */
  public static void main(final String[] args) throws IOException {
    if (args.length > 0 && args.length <= 4) {
      Class<?> clazz = ReflectUtil.loadClass(null, args[0]);
      Constructor<?> c = ReflectUtil.getConstructor(clazz, new Class<?>[]{String.class, InstallationType.class});
      String hibernateRelativePath = null;
      InstallationType installationType = InstallationType.INTERNAL_JOURNAL_WITH_HISTORY;
      File file = null;
      if (args.length > 1 ) {
        file = new File(args[1]);
      }
      if (args.length > 2) {
        hibernateRelativePath = args[2];
      }
      if (args.length > 3) {
        installationType = InstallationType.valueOf(args[3]);
      }
      final DefaultEnvXMLGenerator generator = (DefaultEnvXMLGenerator) ReflectUtil.newInstance(c,
              new Object[]{hibernateRelativePath , installationType});
      String envAsString = generator.createEnvironmentXml();
      if (file == null) {
        System.out.println(envAsString);
      } else {
        Misc.write(envAsString, file);
      }
    } else {
      throw new OrchestraRuntimeException("Invalid arguments");
    }
  }
}
