/*
 * Copyright (C) 2011  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 */
package org.ow2.orchestra.env.descriptor;

import org.hibernate.SessionFactory;
import org.hibernate.ejb.EntityManagerFactoryImpl;
import org.hibernate.ejb.HibernatePersistence;
import org.hibernate.jmx.StatisticsService;
import org.ow2.orchestra.env.WireContext;
import org.ow2.orchestra.env.WireDefinition;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.Listener;
import org.ow2.orchestra.util.Misc;
import org.ow2.orchestra.util.OrchestraConstants;

import javax.management.ObjectName;
import java.lang.management.ManagementFactory;
import java.util.Hashtable;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * @author Guillaume Porcher
 */
public class HibernateEntityManagerFactoryDescriptor extends AbstractDescriptor {

  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(HibernateEntityManagerFactoryDescriptor.class.getName());

  private String persistenceUnitName;
  private PropertiesDescriptor propertiesDescriptor;

  public Object construct(final WireContext wireContext) {
    final Properties properties;
    if (this.propertiesDescriptor != null) {
      properties = (Properties) wireContext.create(
              this.propertiesDescriptor, false);
      Misc.fastDynamicLog(HibernateEntityManagerFactoryDescriptor.LOG, Level.FINE,
              "adding properties to hibernate configuration %s: %s", this.name, properties);
    } else {
      properties = null;
    }
    EntityManagerFactoryImpl entityManagerFactory =
            (EntityManagerFactoryImpl) new HibernatePersistence().createEntityManagerFactory(this.persistenceUnitName, properties);

    wireContext.addListener(new SessionFactoryCloser(entityManagerFactory.getSessionFactory()));

    final Properties orchestraProperties = (Properties) wireContext.get(OrchestraConstants.ORCHESTRA_PROPERTIES);
    if (orchestraProperties != null) {
      final String jmxObjectName = orchestraProperties.getProperty(OrchestraConstants.JMX_OBJECT_NAME);
      final String domainName = jmxObjectName.substring(0, jmxObjectName.indexOf(":"));

      // Construct mbean name from sessionFactoryName
      String mbeanName = this.getName();
      mbeanName = mbeanName.replace(",", "_");
      mbeanName = mbeanName.replace("=", "_");
      mbeanName = mbeanName.replace(":", "_");
      mbeanName = mbeanName.replace("'", "_");
      mbeanName = mbeanName.replace("*", "_");
      mbeanName = mbeanName.replace("?", "_");
      // Register mbean
      final Hashtable<String, String> tb = new Hashtable<String, String>();
      tb.put("type", "Hibernate");
      tb.put("name", mbeanName);
      final ObjectName on;
      try {
        on = new ObjectName(domainName, tb);
        final StatisticsService stats = new StatisticsService();
        stats.setSessionFactory(entityManagerFactory.getSessionFactory());
        ManagementFactory.getPlatformMBeanServer().registerMBean(stats, on);
      } catch (final Exception e) {
        throw new OrchestraRuntimeException("couldn't register mbeans", e);
      }

      wireContext.addListener(new SessionFactoryMBeanCloser(on));
    }
    return entityManagerFactory;
  }

  public static class SessionFactoryMBeanCloser implements Listener {
    private final ObjectName objectName;

    public SessionFactoryMBeanCloser(final ObjectName on) {
      this.objectName = on;
    }

    public void event(final Object source, final String eventName, final Object info) {
      if (WireContext.EVENT_CLOSE.equals(eventName)) {
        try {
          ManagementFactory.getPlatformMBeanServer().unregisterMBean(this.objectName);
        } catch (final Exception e) {
          throw new OrchestraRuntimeException("Exception while unregistering mbean", e);
        }
      }
    }
  }

  public static class SessionFactoryCloser implements Listener {
    private final SessionFactory sessionFactory;

    public SessionFactoryCloser(final SessionFactory sessionFactory) {
      this.sessionFactory = sessionFactory;
    }

    public void event(final Object source, final String eventName, final Object info) {
      if (WireContext.EVENT_CLOSE.equals(eventName)) {
        Misc.fastDynamicLog(HibernateEntityManagerFactoryDescriptor.LOG, Level.FINE, "closing hibernate session factory");
        this.sessionFactory.close();
      }
    }
  }

  @Override
  public Class< ? > getType(final WireDefinition wireDefinition) {
    return EntityManagerFactoryImpl.class;
  }

  public void setPersistenceUnitName(String persistenceUnitName) {
    this.persistenceUnitName = persistenceUnitName;
  }

  public void setPropertiesDescriptor(PropertiesDescriptor propertiesDescriptor) {
    this.propertiesDescriptor = propertiesDescriptor;
  }
}
