/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.jmx.commands;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.extension.ExtensionProvider;
import org.ow2.orchestra.services.commands.Command;
import org.ow2.orchestra.services.itf.Publisher;
import org.ow2.orchestra.services.itf.Repository;
import org.ow2.orchestra.services.job.ExecuteInvokeJob;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.Misc;

import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Command executed during Orchestra startup.
 *
 * This initializes the engine (redeploys web services, extensions...)
 */
public final class InitializeCommand implements Command<Object> {

  private static final Logger LOG = Logger.getLogger(InitializeCommand.class.getName());
  private static final long serialVersionUID = 1906399701851091793L;

  public Object execute(final Environment environment) {
    final Repository repository = environment.get(Repository.class);
    final Publisher publisher = environment.get(Publisher.class);
    Misc.fastDynamicLog(InitializeCommand.LOG, Level.INFO, "Orchestra initialization.");
    final List<ExtensionProvider> extensionProviders = EnvTool.get(List.class, "extensionProviders");
    // For each process:
    // - first deploy web services using the publisher
    // - deploy extensions using extension providers
    for (final BpelProcess process : repository.getProcesses()) {
      Misc.fastDynamicLog(InitializeCommand.LOG, Level.INFO, "Redeploying services for process %s", process.getQName());
      publisher.publishServices(process, environment);
      for (ExtensionProvider extensionProvider : extensionProviders) {
        extensionProvider.deployExternalModule(process);
      }
    }
    // Set running invoke jobs to an error state because the invoke has stopped in an unknown state.
    Misc.fastDynamicLog(InitializeCommand.LOG, Level.INFO, "Fixing running invoke jobs.");
    for (final ExecuteInvokeJob job : repository.getLockedExecuteInvokeJobs()) {
      job.setLockExpirationTime(null);
      job.setLockOwner(null);
      job.setRetries(0);
      job.setException(
              "Invoke was running during server shutdown. "
              + "The status of the remote call is unknown. "
              + "Please check the status before retrying or exiting this instance.");
    }
    Misc.fastDynamicLog(InitializeCommand.LOG, Level.INFO, "Orchestra initialized.");
    return null;
  }
}
