/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.activity.AbstractBpelActivity;
import org.ow2.orchestra.definition.activity.Flow;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.Link;
import org.ow2.orchestra.definition.element.Source;
import org.ow2.orchestra.definition.element.Target;
import org.ow2.orchestra.exception.StaticAnalysisException;
import org.ow2.orchestra.extension.ExtensionProvider;
import org.ow2.orchestra.facade.def.full.BaseFCTFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityWithChildrenFullDefinition;
import org.ow2.orchestra.facade.def.full.BpelActivityWithSingleChildFullDefinition;
import org.ow2.orchestra.facade.def.full.ElseFullDefinition;
import org.ow2.orchestra.facade.def.full.ElseIfFullDefinition;
import org.ow2.orchestra.facade.def.full.OnAlarmFullDefinition;
import org.ow2.orchestra.facade.def.full.OnEventFullDefinition;
import org.ow2.orchestra.facade.def.full.OnMessageFullDefinition;
import org.ow2.orchestra.facade.def.full.impl.SourceFullImpl;
import org.ow2.orchestra.facade.def.full.impl.TargetFullImpl;
import org.ow2.orchestra.facade.def.full.impl.UnknownActivityFullDefinitionImpl;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.facade.uuid.ActivityDefinitionUUID;
import org.ow2.orchestra.facade.uuid.ProcessDefinitionUUID;
import org.ow2.orchestra.lang.Expression;
import org.ow2.orchestra.lang.JoinCondition;
import org.ow2.orchestra.lang.evaluator.EvaluatorFactory;
import org.ow2.orchestra.parsing.binding.extension.Extension;
import org.ow2.orchestra.parsing.binding.extension.ExtensionsInfo;
import org.ow2.orchestra.util.*;
import org.ow2.orchestra.xml.Parse;
import org.ow2.orchestra.xml.TagBinding;
import org.w3c.dom.Attr;
import org.w3c.dom.Element;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Charles Souillard
 */
public abstract class ActivityBinding<T> extends TagBinding<T> {

   private final ActivityType activityType;

    protected ActivityBinding(final String tagName, final String namespaceUri, final String category , ActivityType activityType)  {
        super(tagName, namespaceUri, category);
        this.activityType=activityType;
    }

    /**
     * Set enclosing scope on the given activity.
     *
     * @param activity activity currently analysed
     * @param parse    parse
     */
    public void setEnclosingScope(final AbstractBpelActivity activity, final Parse parse) {
        final Scope enclosingScope = parse.findObject(Scope.class);
        if (enclosingScope != null) {
            activity.setEnclosingScope(enclosingScope);
        }
    }

    /**
     * Parses activities standard attributes.
     *
     * @param activityElement activity dom element to parse
     * @param activity        activity runtime object on which attributes must be set
     * @param parse           parse
     */
    protected void parseStandardAttributes(final Element activityElement, final AbstractBpelActivity activity, final Parse parse) {
        String name = XmlUtil.attribute(activityElement, "name");
        if (name == null) {
            name = XmlConstants.BPEL_DEFAULT_ACTIVITY_NAME;
        }
        activity.setName(name);
        final String suppressJoinFailureSt = XmlUtil.attribute(activityElement, "suppressJoinFailure");
        final AbstractBpelActivity enclosingActivity = parse.findObject(AbstractBpelActivity.class);
        boolean defaultSuppressJoinFailure = false;
        if (enclosingActivity != null) {
            defaultSuppressJoinFailure = enclosingActivity.isSuppressJoinFailure();
        }
        final boolean suppressJoinFailure = BpelUtil.booleanValue(suppressJoinFailureSt, defaultSuppressJoinFailure);
        activity.setSuppressJoinFailure(suppressJoinFailure);

        final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);

        final ProcessDefinitionUUID processDefinitionUUID = definitionKeeper.getProcessFullDefinition().getUUID();

        ActivityDefinitionUUID activityDefinitionUUID = null;
        if (activity.getType().equals(ActivityType.PROCESS)) {
            activityDefinitionUUID = processDefinitionUUID;
        } else {
            activityDefinitionUUID = new ActivityDefinitionUUID(processDefinitionUUID,
                    definitionKeeper.getNextActivityCount(), this.activityType, name);
        }
        activity.setUUID(activityDefinitionUUID);

        final BpelActivityFullDefinition activityFullDefinition =
                this.createActivityFullDefinition(processDefinitionUUID, activityDefinitionUUID, name, parse);
        activityFullDefinition.setSuppressJoinFailure(suppressJoinFailure);
        final Object parentObject = definitionKeeper.getParent();
        // if parent is null, we are on process activity => don't add new activity
        if (parentObject != null) {
            if (parentObject instanceof BpelActivityWithChildrenFullDefinition) {
                final BpelActivityWithChildrenFullDefinition activityWithChildrenFullDefinition =
                        (BpelActivityWithChildrenFullDefinition) parentObject;
                activityWithChildrenFullDefinition.addEnclosedActivity(activityFullDefinition);
            } else if (parentObject instanceof BpelActivityWithSingleChildFullDefinition) {
                final BpelActivityWithSingleChildFullDefinition activityWithSingleChildFullDefinition =
                        (BpelActivityWithSingleChildFullDefinition) parentObject;
                activityWithSingleChildFullDefinition.setEnclosedActivity(activityFullDefinition);
            } else if (parentObject instanceof ElseIfFullDefinition) {
                ((ElseIfFullDefinition) parentObject).setEnclosedActivityFullDefinition(activityFullDefinition);
                ((ElseIfFullDefinition) parentObject).setActivityDefinitionUUID(activityFullDefinition.getUUID());
            } else if (parentObject instanceof ElseFullDefinition) {
                ((ElseFullDefinition) parentObject).setEnclosedActivityFullDefinition(activityFullDefinition);
                ((ElseFullDefinition) parentObject).setActivityDefinitionUUID(activityFullDefinition.getUUID());
            } else if (parentObject instanceof OnMessageFullDefinition) {
                ((OnMessageFullDefinition) parentObject).setEnclosedActivityFullDefinition(activityFullDefinition);
                ((OnMessageFullDefinition) parentObject).setActivityDefinitionUUID(activityFullDefinition.getUUID());
            } else if (parentObject instanceof OnEventFullDefinition) {
                ((OnEventFullDefinition) parentObject).setEnclosedActivityFullDefinition(activityFullDefinition);
                ((OnEventFullDefinition) parentObject).setActivityDefinitionUUID(activityFullDefinition.getUUID());
            } else if (parentObject instanceof OnAlarmFullDefinition) {
                ((OnAlarmFullDefinition) parentObject).setEnclosedActivityFullDefinition(activityFullDefinition);
                ((OnAlarmFullDefinition) parentObject).setActivityDefinitionUUID(activityFullDefinition.getUUID());
            } else if (parentObject instanceof BaseFCTFullDefinition) {
                ((BaseFCTFullDefinition) parentObject).setEnclosedActivity(activityFullDefinition);
            } else {
                Misc.unreachableStatement("parent has bad type " + parentObject.getClass() + " in activity " + activity.getType());
            }
        }
        definitionKeeper.setActivityFullDefinition(activityFullDefinition);
    }

    public BpelActivityFullDefinition createActivityFullDefinition(final ProcessDefinitionUUID processDefinitionUUID,
                                                                   final ActivityDefinitionUUID activityDefinitionUUID, final String name, final Parse parse) {
        return new UnknownActivityFullDefinitionImpl(processDefinitionUUID, activityDefinitionUUID, name);
    }

    /**
     * Parses activities extension attributes.
     *
     * @param activityElement activity dom element to parse
     * @param activity        activity runtime object on which attributes must be set
     * @param parse           parse
     */
    protected static void parseExtensionAttributes(final Element activityElement, final AbstractBpelActivity activity, final Parse parse) {
        final ExtensionsInfo extensionsInfo = parse.findObject(ExtensionsInfo.class);
        List<Attr> nonStandardAttributes = XmlUtil.attributes(activityElement);
        if (nonStandardAttributes != null) {
            for (Attr att : nonStandardAttributes) {
                if (att.getNamespaceURI() != null) {
                    if (!att.getNamespaceURI().equals(XmlConstants.NS_XMLNS)) {
                        if (!att.getNamespaceURI().equals(XmlConstants.NS_XML_SCHEMA_INSTANCE)) {
                            // this is a extension attribute
                            Extension extDef = extensionsInfo.getExtensionDefinitionByNamespace(att.getNamespaceURI());
                            if (extDef == null) {
                                parse.addProblem("Unknown extension attribute: " + att.getName(), activityElement);
                            } else if (!extDef.isSupported()) {
                                if (extDef.isMustUnderstand()) {
                                    parse.addProblem("Unsupported extension attribute: " + att.getName()
                                            + " (extension namespace: " + extDef.getNamespace() + ")", activityElement);
                                } else {
                                    parse.addWarning("Unsupported optional extension attribute: " + att.getName()
                                            + " (extension namespace: " + extDef.getNamespace() + "), it will be ignored.", activityElement);
                                }
                            } else {
                                // Orchestra supported this extension
                                ExtensionProvider extensionProvider = extDef.getExtensionProvider();
                                if (extensionProvider != null) {
                                    extensionProvider.parseExtensionAttribute(activityElement, parse);
                                }
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Parse activities standard elements.
     *
     * @param activityElement activity dom element to parse
     * @param activity        activity runtime object on which elements must be set
     * @param parse           parse stack
     */
    protected void parseStandardElements(final Element activityElement, final AbstractBpelActivity activity, final Parse parse) {
        final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
        final BpelActivityFullDefinition activityFullDefinition =
                (BpelActivityFullDefinition) definitionKeeper.getActivityFullDefinition();

        // sources
        final Element sourcesElement = XmlUtil.element(activityElement, "sources");
        if (sourcesElement != null) {
            final List<Element> sourceElements = XmlUtil.elements(sourcesElement, "source");
            this.parseSources(sourceElements, parse, activity);
        }

        // targets
        final Element targetsElement = XmlUtil.element(activityElement, "targets");
        if (targetsElement != null) {
            final List<Element> targetElements = XmlUtil.elements(targetsElement, "target");
            this.parseTargets(targetElements, parse, activity);

            // targets joinCondition
            final Element targetsJoinConditionElement = XmlUtil.element(targetsElement, "joinCondition");
            String expressionLanguage = null;
            String booleanExpression = null;
            if (targetsJoinConditionElement != null) {
                expressionLanguage = XmlUtil.attribute(targetsJoinConditionElement, "expressionLanguage");
                final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
                if (expressionLanguage == null) {
                    expressionLanguage = bpelProcess.getExpressionLanguage();
                }
                final JoinCondition joinCondition = new JoinCondition();
                booleanExpression = targetsJoinConditionElement.getTextContent();
                joinCondition.setText(booleanExpression);
                if (!EvaluatorFactory.isSupportedLanguage(expressionLanguage)) {
                    // SA00004
                    final String message = StaticAnalysisFault.getMessage("SA00004",
                            "  Expression language '" + expressionLanguage + "' not supported.");
                    parse.addProblem(message, new StaticAnalysisException("SA00004"), targetsJoinConditionElement);
                }
                joinCondition.setLanguage(expressionLanguage);
                joinCondition.setNamespaces(BpelUtil.getAllNameSpaces(targetsJoinConditionElement));
                activity.setTargetsJoinCondition(joinCondition);
            }
            activityFullDefinition.setTargetsJoinConditionExpressionLanguage(expressionLanguage);
            activityFullDefinition.setTargetsJoinConditionBooleanExpression(booleanExpression);
        }
        // documentation
        final List<Element> documentationElements = XmlUtil.elements(activityElement, "documentation");
        if (documentationElements != null) {
            final StringBuilder documentation = new StringBuilder();
            for (final Element documentationElement : documentationElements) {
                if (documentation.length() != 0) {
                    documentation.append(Misc.LINE_SEPARATOR);
                }
                documentation.append(XmlUtil.getContentText(documentationElement));
            }
            activityFullDefinition.setDocumentation(documentation.toString());
        }
    }

    /**
     * Parse sources standard elements.
     *
     * @param sourceElements this activity sources element list
     * @param activity       activity runtime object on which elements must be set
     * @param parse          parse stack
     */
    private void parseSources(final List<Element> sourceElements, final Parse parse, final AbstractBpelActivity activity) {
        final List<Source> sources = new ArrayList<Source>();
        if (sourceElements != null) {
            final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
            final BpelActivityFullDefinition activityFullDefinition =
                    (BpelActivityFullDefinition) definitionKeeper.getActivityFullDefinition();
            final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
            for (final Element sourceElement : sourceElements) {
                final String linkName = XmlUtil.attribute(sourceElement, "linkName");
                boolean exist = false;
                for (final Source source : sources) {
                    if (source.getLinkName().equals(linkName)) {
                        // Static Analysis SA00068
                        final String message = StaticAnalysisFault.getMessage("SA00068",
                                "  Activity " + activity.getName() + " defines 2 sources with same link "
                                        + linkName);
                        // TODO: Annotates attributes with line and column to have a more
                        // precise error indication
                        // parse.addProblem(message, new StaticAnalysisException("SA00068"),
                        // sourceElement.getAttributeNode("linkName"));
                        parse.addProblem(message, new StaticAnalysisException("SA00068"), sourceElement);
                        exist = true;
                        break;
                    }
                }
                if (!exist) {
                    final Source source = new Source();
                    source.setLinkName(linkName);
                    final Element transitionConditionElement = XmlUtil.element(sourceElement, "transitionCondition");
                    String expressionLanguage = null;
                    String booleanExpression = null;
                    if (transitionConditionElement != null) {
                        expressionLanguage = XmlUtil.attribute(transitionConditionElement, "expressionLanguage");
                        if (expressionLanguage == null) {
                            expressionLanguage = bpelProcess.getExpressionLanguage();
                        }
                        final Expression expression = new Expression();
                        if (!EvaluatorFactory.isSupportedLanguage(expressionLanguage)) {
                            // SA00004
                            final String message = StaticAnalysisFault.getMessage("SA00004",
                                    "  Expression language '" + expressionLanguage + "' not supported.");
                            parse.addProblem(message, new StaticAnalysisException("SA00004"), transitionConditionElement);
                        }
                        expression.setLanguage(expressionLanguage);
                        expression.setNamespaces(BpelUtil.getAllNameSpaces(transitionConditionElement));
                        booleanExpression = transitionConditionElement.getTextContent();
                        expression.setText(booleanExpression);
                        source.setTransitionCondition(expression);
                    }
                    sources.add(source);
                    final List<Flow> flows = parse.findObjects(Flow.class);
                    Link link = null;
                    for (final Flow flow : flows) {
                        link = flow.getLink(linkName);
                        if (link != null) {
                            break;
                        }
                    }
                    if (link == null) {
                        // Static Analysis SA00065
                        final String message = StaticAnalysisFault.getMessage("SA00065",
                                "  Activity " + activity.getName() + " defines source with link " + linkName
                                        + " that is not defined in enclosing flow");
                        // TODO: Annotates attributes with line and column to have a more
                        // precise error indication
                        // parse.addProblem(message, new StaticAnalysisException("SA00065"),
                        // sourceElement.getAttributeNode("linkName") == null ?
                        // sourceElement: sourceElement.getAttributeNode("linkName"));
                        parse.addProblem(message, new StaticAnalysisException("SA00065"), sourceElement);
                    } else {
                        if (link.getSourceActivity() == null) {
                            link.setSourceActivity(activity);
                            source.setLink(link);
                        } else {
                            // Static Analysis SA00066
                            final String message = StaticAnalysisFault.getMessage("SA00066", "  Link " + linkName + " has 2 sources defined");
                            // TODO: Annotates attributes with line and column to have a more
                            // precise error indication
                            // parse.addProblem(message, new
                            // StaticAnalysisException("SA00066"),
                            // sourceElement.getAttributeNode("linkName") == null ?
                            // sourceElement
                            // : sourceElement.getAttributeNode("linkName"));
                            parse.addProblem(message, new StaticAnalysisException("SA00066"), sourceElement);
                        }
                    }
                    activityFullDefinition.addSource(new SourceFullImpl(linkName, expressionLanguage, booleanExpression));
                }
            }
        }
        activity.setSources(sources);
    }

    /**
     * Parse targets standard elements.
     *
     * @param targetElements this activity targets element list
     * @param parse          parse
     * @param activity       activity runtime object on which elements must be set
     */
    private void parseTargets(final List<Element> targetElements, final Parse parse, final AbstractBpelActivity activity) {
        final List<Target> targets = new ArrayList<Target>();
        if (targetElements != null) {
            final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
            final BpelActivityFullDefinition activityFullDefinition =
                    (BpelActivityFullDefinition) definitionKeeper.getActivityFullDefinition();
            for (final Element targetElement : targetElements) {
                final String linkName = XmlUtil.attribute(targetElement, "linkName");
                boolean exist = false;
                for (final Target target : targets) {
                    if (target.getLinkName().equals(linkName)) {
                        // Static Analysis SA00069
                        final String message = StaticAnalysisFault.getMessage("SA00069",
                                "  Activity " + activity.getName() + " defines 2 targets with same link "
                                        + linkName);
                        // TODO: Annotates attributes with line and column to have a more
                        // precise error indication
                        // parse.addProblem(message, new StaticAnalysisException("SA00069"),
                        // targetElement.getAttributeNode("linkName"));
                        parse.addProblem(message, new StaticAnalysisException("SA00069"), targetElement);
                        exist = true;
                        break;
                    }
                }
                if (!exist) {
                    final Target target = new Target();
                    target.setLinkName(linkName);
                    targets.add(target);
                    final List<Flow> flows = parse.findObjects(Flow.class);
                    Link link = null;
                    for (final Flow flow : flows) {
                        link = flow.getLink(linkName);
                        if (link != null) {
                            break;
                        }
                    }
                    if (link == null) {
                        // Static Analysis SA00065
                        final String message = StaticAnalysisFault.getMessage("SA00065",
                                "  Activity " + activity.getName() + " defines target with link " + linkName
                                        + " that is not defined in enclosing flow");
                        // TODO: Annotates attributes with line and column to have a more
                        // precise error indication
                        // parse.addProblem(message, new StaticAnalysisException("SA00065"),
                        // targetElement.getAttributeNode("linkName") == null ?
                        // targetElement
                        // : targetElement.getAttributeNode("linkName"));
                        parse.addProblem(message, new StaticAnalysisException("SA00065"), targetElement);
                    } else {
                        if (link.getTargetActivity() == null) {
                            link.setTargetActivity(activity);
                            target.setLink(link);
                        } else {
                            // Static Analysis SA00066
                            final String message = StaticAnalysisFault.getMessage("SA00066", "  Link " + linkName + " has 2 targets defined");
                            // TODO: Annotates attributes with line and column to have a more
                            // precise error indication
                            // parse.addProblem(message, new
                            // StaticAnalysisException("SA00066"),
                            // targetElement.getAttributeNode("linkName") == null ?
                            // targetElement
                            // : targetElement.getAttributeNode("linkName"));
                            parse.addProblem(message, new StaticAnalysisException("SA00066"), targetElement);
                        }
                    }
                    activityFullDefinition.addTarget(new TargetFullImpl(linkName));
                }
            }
        }
        activity.setTargets(targets);
    }

    /**
     * Parse activities extensions elements
     *
     * @param activityElement activity dom element to parse
     * @param activity        activity runtime object on which elements must be set
     * @param parse           parse stack
     */
    protected static void parseExtensionElements(final Element activityElement, final AbstractBpelActivity activity, final Parse parse) {
        final ExtensionsInfo extensionsInfo = parse.findObject(ExtensionsInfo.class);
        final List<Element> nonStandardElements = XmlUtil.elements(activityElement);
        if (nonStandardElements != null) {
            for (Element e : nonStandardElements) {
                if (!e.getNamespaceURI().equals(XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE)) {
                    ActivityBinding.parseExtensionElement(e, extensionsInfo, activity, parse);
                }
            }
        }
    }

    /**
     * Parses an extension element
     *
     * @param element         : extension element of this activity
     * @param extensionsInfos ExtensionsInfo
     */

    private static void parseExtensionElement(Element element, ExtensionsInfo extensionsInfos, final AbstractBpelActivity activity, Parse parse) {
        Extension extDef = extensionsInfos.getExtensionDefinitionByNamespace(element.getNamespaceURI());
        if (extDef == null) {
            parse.addProblem("Unknown extension element: " + element.getTagName(), element);
        } else if (!extDef.isSupported()) {
            if (extDef.isMustUnderstand()) {
                parse.addProblem("Unsupported extension element: " + element.getTagName()
                        + " (extension namespace: " + extDef.getNamespace() + ")", element);
            } else {
                parse.addWarning("Unsupported optional extension element: " + element.getTagName()
                        + " (extension namespace: " + extDef.getNamespace() + "), it will be ignored.", element);
            }
        } else {
            //this extension is supported
            ExtensionProvider extensionProvider = extDef.getExtensionProvider();
            if (extensionProvider != null) {
                extensionProvider.parseExtensionElement(element, parse);
            }
        }
    }

}
