/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.parsing.binding;

import org.ow2.orchestra.definition.BpelProcess;
import org.ow2.orchestra.definition.Node;
import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.activity.Scope;
import org.ow2.orchestra.definition.element.PartnerLink;
import org.ow2.orchestra.facade.def.full.ElseFullDefinition;
import org.ow2.orchestra.facade.def.full.ElseIfFullDefinition;
import org.ow2.orchestra.facade.def.full.OnAlarmFullDefinition;
import org.ow2.orchestra.facade.def.full.OnEventFullDefinition;
import org.ow2.orchestra.facade.def.full.OnMessageFullDefinition;
import org.ow2.orchestra.facade.runtime.ActivityType;
import org.ow2.orchestra.parsing.BpelParser;
import org.ow2.orchestra.util.*;
import org.ow2.orchestra.xml.Parse;
import org.ow2.orchestra.xml.Parser;
import org.w3c.dom.Element;

import javax.xml.namespace.QName;
import java.util.List;
import java.util.Set;

/**
 * @author Charles Souillard
 */
public abstract class BpelActivityBinding extends ActivityBinding<Node> {

    protected BpelActivityBinding(final String tagName, final ActivityType activityType) {
        super(tagName, XmlConstants.XMLNS_BPEL_2_0_EXECUTABLE, BpelParser.CATEGORY_ACTIVITY, activityType);
    }

    /**
     * Takes all children activities of currentActivity and for each one creates a
     * new Node and a new behaviour (parses the activity).
     *
     * @param activityElement dom element representing the currentActivity
     * @param parse           parse stack
     * @param parser          parser
     */
    private void parseChildrenActivities(final Element activityElement, final Parse parse, final Parser parser) {
        final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
        try {
            definitionKeeper.pushParent();

            final Set<QName> allActivityTagNames = BpelParser.getAllActivityQNames();
            final List<Element> activityElements = XmlUtil.elementsQName(activityElement, allActivityTagNames);
            if (activityElements == null || activityElements.size() == 0) {
                parse.addProblem("No activity was specified under this element. Please specify one of "
                        + allActivityTagNames, activityElement);
                return;
            }
            for (final Element childrenElement : activityElements) {
                parser.parseElement(childrenElement, parse, BpelParser.CATEGORY_ACTIVITY);

            }
        } finally {
            definitionKeeper.popParent();
        }
    }

    protected Node parseChildActivity(final Element parentElement, final Parse parse, final Parser parser) {
        final DefinitionKeeper definitionKeeper = parse.findObject(DefinitionKeeper.class);
        try {
            if (!(definitionKeeper.getParent() instanceof ElseIfFullDefinition)
                    && !(definitionKeeper.getParent() instanceof ElseFullDefinition)
                    && !(definitionKeeper.getParent() instanceof OnMessageFullDefinition)
                    && !(definitionKeeper.getParent() instanceof OnEventFullDefinition)
                    && !(definitionKeeper.getParent() instanceof OnAlarmFullDefinition)) {
                definitionKeeper.pushParent();
            }
            final Set<QName> allActivityTagNames = BpelParser.getAllActivityQNames();
            final List<Element> activityElements = XmlUtil.elementsQName(parentElement, allActivityTagNames);
            if (activityElements.size() == 0) {
                parse.addProblem("No activity was specified under this element. Please specify one of "
                        + allActivityTagNames, parentElement);
                return null;
            }
            final Element childrenElement = activityElements.get(0);
            return (Node) parser.parseElement(childrenElement, parse, BpelParser.CATEGORY_ACTIVITY);
        } finally {
            if (!(definitionKeeper.getParent() instanceof ElseIfFullDefinition)
                    && !(definitionKeeper.getParent() instanceof ElseFullDefinition)
                    && !(definitionKeeper.getParent() instanceof OnMessageFullDefinition)
                    && !(definitionKeeper.getParent() instanceof OnEventFullDefinition)
                    && !(definitionKeeper.getParent() instanceof OnAlarmFullDefinition)) {
                definitionKeeper.popParent();
            }
        }
    }

    public Node createActivityNodeWithChildren(final Element activityElement,
                                               final Parse parse, final Parser parser, final AbstractActivity activityBehaviour) {
        final Node activityNode = this.createActivityNode(parse, parser, activityBehaviour);
        parse.pushObject(activityNode);
        parse.pushObject(activityBehaviour);
        try {
            this.parseChildrenActivities(activityElement, parse, parser);
        } finally {
            parse.popObject();
            parse.popObject();
        }
        return activityNode;
    }

    protected String getNodeName(final Element activityElement) {
        return XmlUtil.getTagLocalName(activityElement) + " " + XmlUtil.attribute(activityElement, "name");
    }

    protected String getNodeName(final String activityNodeName, final String activityName) {
        return activityNodeName + " " + activityName;
    }

    protected Node createActivityNode(final Parse parse, final Parser parser, final AbstractActivity activityBehaviour) {
        final Node parentNode = parse.findObject(Node.class);
        Node activityNode = null;
        if (parentNode != null) {
            activityNode = parentNode.createNode(activityBehaviour.getUUID());
        } else {
            final BpelProcess bpelProcess = parse.findObject(BpelProcess.class);
            activityNode = bpelProcess.createNode(activityBehaviour.getUUID());
        }
        activityNode.setBehaviour(activityBehaviour);
        return activityNode;
    }

    /**
     * Takes the current scope definition and look into for a partnerLink having
     * the given name. If not, look into scope father...
     *
     * @param partnerLinkName partnerLinkName of the searched PartnerLink
     * @param parse           parse stack
     * @return the node corresponding to the currentActivity
     */
    protected PartnerLink findPartnerLinkDefinition(final Parse parse, final String partnerLinkName) {
        final Scope enclosingScope = parse.findObject(Scope.class);
        return enclosingScope.findPartnerLink(partnerLinkName);
    }
}
