/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.persistence.tx;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.transaction.Synchronization;

import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.util.Misc;

/**
 * simple 2 phase commit transaction. no logging or recovery. non thread safe
 * (which is ok).
 *
 * @author Tom Baeyens
 */
public class StandardTransaction implements Transaction, Serializable {

  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(StandardTransaction.class.getName());

  static enum State {
    CREATED, ACTIVE, ROLLBACKONLY, COMMITTED, ROLLEDBACK
  }

  protected final List<StandardResource> resources = new LinkedList<StandardResource>();
  protected final List<StandardSynchronization> synchronizations = new ArrayList<StandardSynchronization>();
  protected State state = State.CREATED;

  // methods for interceptor //////////////////////////////////////////////////

  public void begin() {
    Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "beginning %s", this);
    this.state = State.ACTIVE;
  }

  public void complete() {
    if (this.state == State.ACTIVE) {
      this.commit();
    } else if (this.state == State.ROLLBACKONLY) {
      this.rollback();
    } else {
      throw new TransactionException("complete on transaction in state "
          + this.state);
    }
  }

  // public tx methods ////////////////////////////////////////////////////////

  public void setRollbackOnly() {
    if (this.state != State.ACTIVE) {
      throw new TransactionException("transaction was not active: " + this.state);
    }
    this.state = State.ROLLBACKONLY;
  }

  public boolean isRollbackOnly() {
    return (this.state == State.ROLLBACKONLY) || (this.state == State.ROLLEDBACK);
  }

  // commit ///////////////////////////////////////////////////////////////////

  /** implements simplest two phase commit. */
  public void commit() {
    if (this.state != State.ACTIVE) {
      throw new TransactionException("commit on transaction in state " + this.state);
    }

    Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "committing %s", this);

    try {
      this.beforeCompletion();

      if (this.resources != null) {
        // prepare
        // //////////////////////////////////////////////////////////////
        // the prepare loop will be skipped at the first exception
        for (final StandardResource standardResource : this.resources) {
          Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "preparing resource %s", standardResource);
          standardResource.prepare();
        }
      }

      // for any exception in the prepare phase, we'll rollback
    } catch (final RuntimeException exception) {
      try {
        Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINE, "resource threw exception in prepare. rolling back.");
        this.rollbackResources();
      } catch (final Exception rollbackException) {
        Misc.fastDynamicLog(StandardTransaction.LOG, Level.SEVERE, "rollback failed as well: %s %s",
            rollbackException, Misc.getStackTraceFrom(rollbackException));
      }

      // rethrow
      throw exception;
    }

    // here is the point of no return :-)

    // commit ///////////////////////////////////////////////////////////////
    Throwable commitException = null;
    if (this.resources != null) {
      // The commit loop will try to send the commit to every resource,
      // No matter what it takes. If exceptions come out of resource.commit's
      // they will be suppressed and the first exception will be rethrown after
      // all the resources are commited
      for (final StandardResource standardResource : this.resources) {
        try {
          Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "committing resource %s", standardResource);
          standardResource.commit();

          // Exceptions in the commit phase will not lead to rollback, since
          // some resources
          // might have committed and can't go back.
        } catch (final Throwable t) {
          // TODO this should go to a special log for sys admin recovery
          Misc.fastDynamicLog(StandardTransaction.LOG, Level.SEVERE, "commit failed for resource %s: %s %s",
              standardResource, t, Misc.getStackTraceFrom(t));
          if (commitException == null) {
            commitException = t;
          }
        }
      }
    }

    this.state = State.COMMITTED;
    this.afterCompletion();
    Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "committed %s", this);

    if (commitException != null) {
      if (commitException instanceof RuntimeException) {
        throw (RuntimeException) commitException;
      } else if (commitException instanceof Error) {
        throw (Error) commitException;
      }
      throw new TransactionException("resource failed to commit",
          commitException);
    }
  }

  // rollback /////////////////////////////////////////////////////////////////

  public void rollback() {
    if ((this.state != State.ACTIVE) && (this.state != State.ROLLBACKONLY)) {
      throw new TransactionException("rollback on transaction in state "
          + this.state);
    }

    Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "rolling back %s", this);

    this.beforeCompletion();
    this.rollbackResources();
  }

  void rollbackResources() {
    if (this.resources != null) {
      for (final StandardResource resource : this.resources) {
        try {
          Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "rolling back resource %s", resource);
          resource.rollback();
        } catch (final RuntimeException e) {
          Misc.fastDynamicLog(StandardTransaction.LOG, Level.SEVERE, "rollback failed for resource %s: %s %s",
              resource, e, Misc.getStackTraceFrom(e));
        }
      }
    }

    this.state = State.ROLLEDBACK;

    this.afterCompletion();

    Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "rolled back");
  }

  // synchronizations /////////////////////////////////////////////////////////

  public void registerSynchronization(final Synchronization synchronization) {
    this.synchronizations.add(new StandardSynchronization(synchronization));
  }

  public void afterCompletion() {
    if (this.synchronizations != null) {
      for (final StandardSynchronization synchronization : this.synchronizations) {
        synchronization.afterCompletion(this.state);
      }
    }
  }

  public void beforeCompletion() {
    if (this.synchronizations != null) {
      for (final StandardSynchronization synchronization : this.synchronizations) {
        synchronization.beforeCompletion();
      }
    }
  }

  // resource enlisting ///////////////////////////////////////////////////////

  public void enlistResource(final StandardResource standardResource) {
    Misc.fastDynamicLog(StandardTransaction.LOG, Level.FINEST, "enlisting resource %s to standard transaction", standardResource);
    this.resources.add(0, standardResource);
  }

  List<StandardResource> getResources() {
    return this.resources;
  }

  @Override
  public String toString() {
    return "StandardTransaction[" + System.identityHashCode(this) + ", state=" + this.state + "]";
  }
}
