/**
 * Copyright (C) 2006  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.runtime;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;

import org.ow2.orchestra.definition.activity.AbstractActivity;
import org.ow2.orchestra.definition.element.Correlation;
import org.ow2.orchestra.definition.element.Correlation.Initiate;
import org.ow2.orchestra.definition.element.CorrelationSet;
import org.ow2.orchestra.facade.data.runtime.ActivityInstanceData;
import org.ow2.orchestra.facade.data.runtime.CorrelationSetInitializedData;
import org.ow2.orchestra.facade.exception.BpelFaultException;
import org.ow2.orchestra.facade.exception.BpelFaultException.BpelFault;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.util.EnvTool;
import org.ow2.orchestra.util.XmlConstants;
import org.ow2.orchestra.util.XmlUtil;
import org.ow2.orchestra.var.MessageVariable;
import org.ow2.orchestra.wsdl.PropertyAlias;
import org.ow2.orchestra.wsdl.WsdlsInfos;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * @author Charles Souillard, Guillaume Porcher
 */
public class CorrelationSetRuntime {

  protected long dbid;
  protected int dbversion;

  protected static final Logger LOG = Logger.getLogger(CorrelationSetRuntime.class.getName());

  /** cs name. */
  protected String name;

  /** map of runtime values of properties used by this cs. */
  protected Map<QName, Document> propertyValues;

  /** is cs initialized */
  protected boolean initialized = false;

  protected ScopeRuntime scopeRuntime;

  protected CorrelationSetRuntime() {
  }

  /**
   * Default constructor.
   *
   * @param correlationSetDefinition correlationSetDefinition
   */
  public CorrelationSetRuntime(final CorrelationSet correlationSetDefinition,
    final ScopeRuntime scopeRuntime) {
    this.scopeRuntime = scopeRuntime;
    this.name = correlationSetDefinition.getName();
    this.propertyValues = new HashMap<QName, Document>();
  }

  /**
   * Return the runtime value of the property identified by the given
   * propertyName.
   *
   * @param propertyName
   *            propertyName
   * @return the runtime value of the property identified by the given
   *         propertyName.
   */
  public Node getPropertyValue(final QName propertyName) {
    final Document doc = this.propertyValues.get(propertyName);
    if (doc != null) {
      final Element elt = doc.getDocumentElement();
      final Element child = XmlUtil.element(elt);
      if (child != null) {
        return child;
      }
      return elt.getFirstChild();
    }
    return null;
  }

  /**
   * Set the runtime value of the property identified by the given propertyName.
   *
   * @param propertyName
   *            propertyName
   * @param value
   *            value to set
   */
  public void setPropertyValue(final QName propertyName, final Node value) {
    final Element el = XmlUtil.getDocumentWithOneElement(
        new QName(XmlConstants.ELT_DEFAULT_NS, XmlConstants.ELT_DEFAULT_ELEMENT));
    el.appendChild(el.getOwnerDocument().importNode(value, true));
    this.propertyValues.put(propertyName, el.getOwnerDocument());
  }

  protected boolean validatePropertyValue(final PropertyAlias propertyAlias, final MessageVariable message) {
    try {
      final Node expectedValue = this.getPropertyValue(propertyAlias.getPropertyName());

      final Node value = propertyAlias.getPropertyValueFromMessage(message);

      if (value == null) {
        return expectedValue == null;
      } else if (expectedValue == null) {
        return false;
      }
      final boolean valid = XmlUtil.areNodeEquals(expectedValue, value);
      if (CorrelationSetRuntime.LOG.isLoggable(Level.FINEST)) {
        if (!valid) {
          CorrelationSetRuntime.LOG.finest("Trying to validate CS value: expected " + XmlUtil.toString(expectedValue) + " got "
            + XmlUtil.toString(value));
        } else {
          CorrelationSetRuntime.LOG.finest("CS value valid: " + XmlUtil.toString(expectedValue));
        }
      }
      return valid;
    } catch (final Exception ex) {
      throw new OrchestraRuntimeException("In Correlation: Function call exception:" + ex, ex);
    }
  }

  public boolean validateCSValues(final WsdlsInfos wsdlsInfos,
      final MessageVariable message,
      final QName messageType,
      final Correlation correlation) {
    final List<PropertyAlias> propertyAliases = correlation.getCorrelationSet().getPropertyAliasRefs(wsdlsInfos, messageType);

    //foreach propertyAlias, we extract the value from the message.
    for (final PropertyAlias propertyAlias : propertyAliases) {
      if (!this.validatePropertyValue(propertyAlias, message)) {
        return false;
      }
    }
    return true;
  }

  public void checkCSValues(final BpelExecution execution, final MessageVariable message,
      final QName messageType, final Correlation correlation) {
    final Initiate initiate = correlation.getInitiate();
    // initialize CS
    if (Initiate.yes.equals(initiate) || !this.isInitialized()) {
      this.initiate(execution, message, messageType, correlation);
    } else {
      // check CS values
      if (!this.validateCSValues(execution.getProcessDefinition().getWsdlInfos(), message, messageType, correlation)) {
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.correlationViolation,
            "Correlation consistency constraint violated",
            null);
      }
    }
  }

  public void initiate(final BpelExecution execution, final MessageVariable message,
      final QName messageType, final Correlation correlation) {

    final Initiate initiate = correlation.getInitiate();

    if (this.isInitialized()) {
      if (Initiate.yes.equals(initiate)) {
        throw BpelFaultException.generateBpelFaultException(
            BpelFault.correlationViolation,
            "CorrelationSet already initialized and initiate='yes'",
            null);
      }
      return;
    }

    if (Initiate.no.equals(initiate)) {
      throw BpelFaultException.generateBpelFaultException(
          BpelFault.correlationViolation,
          "CorrelationSet not initialized and initiate='no'",
          null);
    }

    final List<PropertyAlias> propertyAliases =
      execution.getProcessDefinition()
               .getWsdlInfos()
               .getPropertyAliases(correlation.getCorrelationSet().getProperties().values(), messageType);
    //TODO : see if we can have an Element type message

    //foreach propertyAlias, we extract the value from the message.
    for (final PropertyAlias propertyAlias : propertyAliases) {
      try {
        final Element partElement = message.getPartValue(propertyAlias.getPart());

        Node value = partElement;
        if (propertyAlias.getQuery() != null) {
          value = propertyAlias.getQuery().getEvaluator().evaluate(partElement);
        }
        this.setPropertyValue(propertyAlias.getPropertyName(), value);
      } catch (final Exception ex) {
        throw new OrchestraRuntimeException("In Correlation: Function call exception:" + ex, ex);
      }
    }
    if (CorrelationSetRuntime.LOG.isLoggable(Level.FINEST)) {
      CorrelationSetRuntime.LOG.finest("Initialized correlation set with values: " + this.propertyValues);
    }
    final AbstractActivity activity = execution.getNode().getBehaviour();
    this.setInitialized(true, activity.createRuntimeData(execution));
  }

  @Override
  public boolean equals(final Object obj) {
    if (obj == null) {
      return false;
    }
    if (!(obj instanceof CorrelationSetRuntime)) {
      return false;
    }
    final CorrelationSetRuntime corr = (CorrelationSetRuntime) obj;
    return this.name.equals(corr.getName())
      && this.propertyValues.equals(corr.getPropertyValues());
  }

  @Override
  public int hashCode() {
    final StringBuffer result = new StringBuffer();
    result.append("name=\"" + this.name + "\" properties=\"");
    for (final QName property : this.propertyValues.keySet()) {
      result.append(property);
    }
    result.append("\"");

    return result.toString().hashCode();
  }

  public Map<QName, Document> getPropertyValues() {
    return this.propertyValues;
  }

  public void setPropertyValues(final Map<QName, Document> propertyValues) {
    this.propertyValues = propertyValues;
  }

  public String getName() {
    return this.name;
  }

  public void setName(final String name) {
    this.name = name;
  }

  public boolean isInitialized() {
    return this.initialized;
  }
  public void setInitialized(final boolean isInitialized, final ActivityInstanceData activityData) {
    this.initialized = isInitialized;
    EnvTool.getRecorder().recordCorrelationSetInitialized(
      new CorrelationSetInitializedData(
          this.scopeRuntime.getActivityInstanceUUID(),
          activityData,
          this.name,
          this.propertyValues));
  }
}
