/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.services.commands;

import org.hibernate.HibernateException;
import org.hibernate.StaleStateException;
import org.hibernate.exception.GenericJDBCException;
import org.hibernate.exception.LockAcquisitionException;
import org.ow2.orchestra.facade.exception.DeploymentException;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.facade.exception.OrchestraWrapperException;
import org.ow2.orchestra.services.jobexecutor.AcquireJobsCmd;
import org.ow2.orchestra.services.jobexecutor.ExecuteJobCmd;
import org.ow2.orchestra.util.ArchiveTool.AtomicArchiveException;
import org.ow2.orchestra.util.Misc;

import java.util.Random;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * retries the command execution in case hibernate throws optimistic locking
 * (StaleObjectException) exceptions.
 *
 * @author Tom Baeyens
 */
public class RetryInterceptor extends Interceptor {

  private static final Logger LOG = Logger.getLogger(RetryInterceptor.class.getName());

  private static final Random RANDOM = new Random();
  private int retries = 3;
  private long delay = 50;
  private long maxDelay = 5000;
  private int delayFactor = 2;

  public <T> T execute(final Command<T> command) {
    int attempt = 1;
    long sleepTime = this.delay;
    long attemptFactor = this.delayFactor;
    while (attempt <= this.retries) {
      if (attempt > 1) {
        Misc.fastDynamicLog(RetryInterceptor.LOG, Level.FINEST, "retrying...");
      }
      try {

        return this.next.execute(command);

      } catch (final RuntimeException e) {
        if (this.shouldRetry(command, e)) {
          Misc.fastDynamicLog(RetryInterceptor.LOG, Level.FINEST,
              "command failed: %s %s", e, Misc.getStackTraceFrom(e));
          attempt++;
          if (attempt > this.retries) {
            Misc.fastDynamicLog(RetryInterceptor.LOG, Level.SEVERE,
                "gave up after %s attempts. Last exception was: %s %s", attempt, e, Misc.getStackTraceFrom(e));
            throw e;
          }
          Misc.fastDynamicLog(RetryInterceptor.LOG, Level.FINEST,
              "waiting %s millis", sleepTime);
          try {
            Thread.sleep(sleepTime);
          } catch (final InterruptedException e1) {
            Misc.fastDynamicLog(RetryInterceptor.LOG, Level.FINEST,
                "retry sleeping got interrupted");
          }
          if (delayFactor > 1) {
            attemptFactor = attemptFactor * (1 + RetryInterceptor.RANDOM.nextInt(this.delayFactor - 1));
          }
          sleepTime = this.delay * attemptFactor;
          if (sleepTime > this.maxDelay) {
            sleepTime = this.maxDelay;
          }
        } else {
          throw e;
        }
      }
    }
    throw new OrchestraRuntimeException("gave up after " + attempt + " attempts");
  }

  protected boolean shouldRetry(final Command<?> command, final RuntimeException e) {
    if (command instanceof ExecuteJobCmd) {
      return false;
    }
    if (command instanceof AcquireJobsCmd) {
      return false;
    }
    if (e instanceof DeploymentException) {
      return false;
    }
    if (e instanceof AtomicArchiveException) {
      return true;
    }
    if (e instanceof OrchestraWrapperException) {
      return false;
    }
    if (e instanceof OrchestraRuntimeException) {
      return true;
    }
    if (e instanceof IllegalStateException) {
      // thrown by misc.unreachableStatement()
      // can be due to a concurrent modification of the journal.
      return true;
    }
    if (e instanceof HibernateException) {
      return true;
    }
    if (e instanceof StaleStateException
        || e instanceof LockAcquisitionException
        || e instanceof GenericJDBCException) {
      return true;
    }
    return false;
  }

  public int getRetries() {
    return this.retries;
  }

  public void setRetries(final int retries) {
    this.retries = retries;
  }

  public long getDelay() {
    return this.delay;
  }

  public void setDelay(final long delay) {
    this.delay = delay;
  }

  public int getDelayFactor() {
    return this.delayFactor;
  }

  public void setDelayFactor(final int delayFactor) {
    this.delayFactor = delayFactor;
  }


  /**
   * @return the maxDelay
   */
  public long getMaxDelay() {
    return this.maxDelay;
  }


  /**
   * @param maxDelay the maxDelay to set
   */
  public void setMaxDelay(final long maxDelay) {
    this.maxDelay = maxDelay;
  }

  /**
   * @return the random
   */
  public static Random getRandom() {
    return RetryInterceptor.RANDOM;
  }
}
