/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.services.job;

import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.facade.exception.OrchestraRuntimeException;
import org.ow2.orchestra.persistence.Transaction;
import org.ow2.orchestra.services.jobexecutor.JobAddedNotification;
import org.ow2.orchestra.services.jobexecutor.JobDbSession;
import org.ow2.orchestra.services.jobexecutor.JobExecutor;
import org.ow2.orchestra.util.Clock;
import org.ow2.orchestra.util.Misc;

/**
 * a runtime timer instance.
 *
 * <p>
 * Timers can be
 * </p>
 *
 * @author Tom Baeyens
 * @author Pascal Verdage
 * @author Alejandro Guizar
 */
public class TimerImpl extends JobImpl<Boolean> implements Timer {

  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(TimerImpl.class.getName());

  private static final String DATE_FORMAT = "yyyy-MM-dd HH:mm:ss,SSS";

  protected String signalName;
  protected Long repeat;

  public static final String EVENT_TIMER = "timer";

  public TimerImpl() {
  }

  public Boolean execute(final Environment environment) {
    Misc.fastDynamicLog(TimerImpl.LOG, Level.FINE, "executing %s", this);

    if (environment == null) {
      throw new OrchestraRuntimeException("environment is null");
    }

    if (this.signalName != null) {
      Misc.fastDynamicLog(TimerImpl.LOG, Level.FINE, "feeding timer signal %s into %s", this.signalName, this.execution);
      this.execution.signal(this.signalName);
    }

    boolean deleteThisJob = true;
    // if there is no repeat on this timer
    if (this.repeat == null) {
      // delete the jobImpl
      Misc.fastDynamicLog(TimerImpl.LOG, Level.FINE, "deleting %s", this);
      final JobDbSession dbSession = environment.get(JobDbSession.class);
      if (dbSession == null) {
        throw new OrchestraRuntimeException("no " + JobDbSession.class.getName()
            + " in environment");
      }
      dbSession.delete(this);

    } else { // there is a repeat on this timer
      deleteThisJob = false;
      // suppose that it took the timer runner thread a very long time to
      // execute the timers
      // then the repeat action dueDate could already have passed
      do {
        this.dueDate = new Date(System.currentTimeMillis() + this.repeat.longValue());
      } while (this.dueDate.getTime() <= Clock.getCurrentTime().getTime());

      Misc.fastDynamicLog(TimerImpl.LOG, Level.FINE, "rescheduled %s for %s", this, TimerImpl.formatDueDate(this.dueDate));

      // release the lock on the timer
      this.setLockOwner(null);
      this.setLockExpirationTime(null);

      // notify the jobExecutor at the end of the transaction
      final JobExecutor jobExecutor = environment.get(JobExecutor.class);
      if (jobExecutor != null) {
        final Transaction transaction = environment.get(Transaction.class);
        if (transaction == null) {
          throw new OrchestraRuntimeException("no transaction in environment");
        }
        final JobAddedNotification jobNotificator = new JobAddedNotification(jobExecutor, this);
        transaction.registerSynchronization(jobNotificator);
      }
    }

    return deleteThisJob;
  }

  @Override
  public String toString() {
    final StringBuffer buffer = new StringBuffer();
    buffer.append("timer[");
    buffer.append(this.dbid);
    if (this.dueDate != null) {
      buffer.append("|");
      buffer.append(TimerImpl.formatDueDate(this.dueDate));
    }
    if (this.signalName != null) {
      buffer.append("|");
      buffer.append(this.signalName);
    }
    buffer.append("]");
    return buffer.toString();
  }

  public static String formatDueDate(final Date date) {
    return new SimpleDateFormat(TimerImpl.DATE_FORMAT).format(date);
  }

  public String getSignalName() {
    return this.signalName;
  }

  public void setSignalName(final String signalName) {
    this.signalName = signalName;
  }

  public Long getRepeat() {
    return this.repeat;
  }

  public void setRepeat(final Long repeat) {
    this.repeat = repeat;
  }
}
