/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.services.jobexecutor;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashSet;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.env.Environment;
import org.ow2.orchestra.runtime.BpelInstance;
import org.ow2.orchestra.services.commands.Command;
import org.ow2.orchestra.services.job.JobImpl;
import org.ow2.orchestra.util.Misc;

/**
 * @author Tom Baeyens
 */
public class AcquireJobsCmd implements Command<Collection<JobImpl<?>>> {

  private static final long serialVersionUID = 1L;
  private static final Logger LOG = Logger.getLogger(AcquireJobsCmd.class.getName());

  private final DateFormat timeFormat = new SimpleDateFormat("HH:mm:ss,SSS");

  private final String lockOwnerName;
  private final int lockMillis;

  private final AbstractJobExecutor jobExecutor;
  private final boolean limitSameInstanceJobs;

  public AcquireJobsCmd(
      final AbstractJobExecutor jobExecutor,
      final String lockOwnerName,
      final int lockMillis,
      final boolean limitSameInstanceJobs) {
    this.jobExecutor = jobExecutor;
    this.lockOwnerName = lockOwnerName;
    this.lockMillis = lockMillis;
    this.limitSameInstanceJobs = limitSameInstanceJobs;
  }

  public Collection<JobImpl<?>> execute(final Environment environment) {
    final Collection<JobImpl<?>> result = new ArrayList<JobImpl<?>>();

    final JobDbSession jobDbSession = environment.get(JobDbSession.class);
    Misc.fastDynamicLog(AcquireJobsCmd.LOG, Level.FINE, "start querying first acquirable jobImpl...");

    final int limit = this.jobExecutor.getAcquireJobSize();
    final Collection<JobImpl< ? >> jobs = jobDbSession.findFirstAcquirableJob(limit, this.limitSameInstanceJobs);

    if (jobs != null && !jobs.isEmpty()) {
      final Collection<BpelInstance> instances = new HashSet<BpelInstance>();
      for (final JobImpl<?> acquiredJob : jobs) {
        if (instances.contains(acquiredJob.getExecution().getProcessInstance())) {
          continue;
        }
        instances.add(acquiredJob.getExecution().getProcessInstance());
        final long lockExpirationTime = System.currentTimeMillis() + this.lockMillis;
        Misc.fastDynamicLog(AcquireJobsCmd.LOG, Level.FINEST, "trying to obtain a lock for '%s' with exp %s",
            acquiredJob, this.timeFormat.format(new Date(lockExpirationTime)));
        acquiredJob.setLockExpirationTime(new Date(lockExpirationTime));
        acquiredJob.setLockOwner(this.lockOwnerName);
        result.add(acquiredJob);
      }
    } else {
      Misc.fastDynamicLog(AcquireJobsCmd.LOG, Level.FINEST, "no acquirable jobs in jobImpl table");
    }

    Misc.fastDynamicLog(AcquireJobsCmd.LOG, Level.FINE, "locking jobs %s", result);

    return result;
  }

}
