/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.ow2.orchestra.services.jobexecutor.pvm;

import java.util.concurrent.RejectedExecutionException;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.ow2.orchestra.services.job.JobImpl;
import org.ow2.orchestra.services.jobexecutor.AbstractDispatcherThread;
import org.ow2.orchestra.services.jobexecutor.jdk.JobExecutorTask;
import org.ow2.orchestra.util.Misc;

/**
 * this thread is responsible for acquiring jobs in the jobImpl that need to be
 * executed and then let the JobExecutor dispatch the acquired ids to one of the
 * JobExecutorThreads in the pool. There is only one dispatcher thread per
 * JobExecutor.
 *
 * @author Tom Baeyens, Guillaume Porcher
 */
public class DispatcherThread extends AbstractDispatcherThread {

  private static final Logger LOG = Logger.getLogger(DispatcherThread.class.getName());

  private final PvmJobExecutor jobExecutor;

  DispatcherThread(final PvmJobExecutor jobExecutor) {
    this(jobExecutor, "DispatcherThread");
  }

  DispatcherThread(final PvmJobExecutor jobExecutor, final String name) {
    super(name);
    this.jobExecutor = jobExecutor;
  }

  @Override
  protected void putAcquiredJobOnQueue(JobImpl<?> acquiredJob) {
    Misc.fastDynamicLog(DispatcherThread.LOG, Level.FINE, "pushing job on the queue %s", acquiredJob);
    while (acquiredJob != null) {
      try {
        final JobExecutorTask task = new JobExecutorTask(this.jobExecutor.getCommandExecutor(), acquiredJob);
        try {
          this.jobExecutor.getThreadPool().execute(task);
        } catch (final RejectedExecutionException e) {
          this.jobExecutor.getJobTasksQueue().put(task);
        }
        Misc.fastDynamicLog(DispatcherThread.LOG, Level.FINEST, "job %s was put on the queue", acquiredJob);
        acquiredJob = null;
      } catch (final InterruptedException e) {
        Misc.fastDynamicLog(DispatcherThread.LOG, Level.FINEST, "putting acquired job got interrupted. retrying...");
      }
    }
  }


  @Override
  public PvmJobExecutor getJobExecutor() {
    return this.jobExecutor;
  }
}
